import sys    
import logging
import pathlib
import time
cur_path = pathlib.Path(__file__).parent.resolve()
sys.path.insert(0, str(cur_path.parent.parent.resolve()))

# Arguments
import argparse
parser = argparse.ArgumentParser()
parser.add_argument('connMethod', type=str, help='Connection method (wifi or usb)')
parser.add_argument('locator', type=str, help = 'IP Address for Wifi or Serial Port for USB', nargs='?')
args = parser.parse_args()

# Import Marty from the martypy library
from martypy import Marty

# Connect to a Marty and# use the variable my_marty to refer to that Marty
my_marty = Marty(args.connMethod, args.locator, blocking=True)

# Setup logging
logging.basicConfig(format='%(levelname)s: %(asctime)s %(funcName)s(%(lineno)d) -- %(message)s', level=logging.DEBUG)
logger = logging.getLogger("testBaudRateSelection")
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)

# Function used for callback
def progressCB(bytesSent, totalBytes, interface):
    print(f"Progress {bytesSent}/{totalBytes}")
    return True

fileName = "completed_tone_low_br.mp3"
# fileName = "unplgivy.mp3"
# fileName = "unplugged.mp3"
# fileName = "test440ToneQuietShort.mp3"
my_marty.client.ricIF.streamSoundFile(cur_path.joinpath(fileName), progressCB, "streamaudio")

# Disconnect from Marty
my_marty.close()


from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Optional, Union, Tuple
from warnings import warn

class ClientGeneric(ABC):

    SIDE_CODES = {
        'left'    : 0,
        'right'   : 1,
        'forward' : 2,
        'back'    : 3,
        'auto'    : 0,
    }

    EYE_POSES = {
        'angry'   : 'eyesAngry',
        'excited' : 'eyesExcited',
        'normal'  : 'eyesNormal',
        'wide'    : 'eyesWide',
        'wiggle'  : 'wiggleEyes'
    }

    DISCO_PATTERNS = {
        "show-off": "show-off",
        "pinwheel": "pinwheel",
        "off": "off"
    }

    NOT_IMPLEMENTED = "Unfortunately this Marty doesn't do that"

    def __init__(self, blocking: Union[bool, None], *args, **kwargs):
        super().__init__()
        if len(args) > 0:
            warn(f"Ignoring unexpected constructor argument(s): {args}", stacklevel=4)
        if len(kwargs) > 0:
            warn(f"Ignoring unexpected constructor argument(s): {kwargs}", stacklevel=4)
        self._is_blocking: bool = True if blocking is None else blocking

    @classmethod
    def dict_merge(cls, *dicts):
        '''
        Merge all provided dicts into one dict
        '''
        merged = {}
        for d in dicts:
            if not isinstance(d, dict):
                raise ValueError('Value should be a dict')
            else:
                merged.update(d)
        return merged

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def close(self):
        pass

    def is_blocking(self, local_override: Optional[bool] = None) -> bool:
        """
        Check if this client is blocking, optionally taking into account a local
        blocking override flag.
        """
        if local_override is not None:
            return local_override
        else:
            return self._is_blocking

    def set_blocking(self, blocking: bool):
        self._is_blocking = blocking

    @abstractmethod
    def wait_if_required(self, expected_wait_ms: int, blocking_override: Union[bool, None]):
        raise NotImplementedError()

    @abstractmethod
    def hello(self) -> bool:
        return False

    @abstractmethod
    def get_ready(self) -> bool:
        return False

    @abstractmethod
    def stand_straight(self, move_time: int) -> bool:
        return False

    @abstractmethod
    def discover(self) -> List[str]:
        return []

    @abstractmethod
    def stop(self, stop_type: str, stopCode: int) -> bool:
        return False

    @abstractmethod
    def resume(self) -> bool:
        return False

    @abstractmethod
    def hold_position(self, hold_time: int) -> bool:
        return False

    @abstractmethod
    def move_joint(self, joint_id: int, position: int, move_time: int) -> bool:
        return False

    @abstractmethod
    def get_joint_position(self, joint_id: Union[int, str]) -> float:
        return 0

    @abstractmethod
    def get_joint_current(self, joint_id: Union[int, str]) -> float:
        return 0

    @abstractmethod
    def get_joint_status(self, joint_id: Union[int, str]) -> int:
        return 0

    @abstractmethod
    def lean(self, direction: str, amount: Optional[int], move_time: int) -> bool:
        return False

    @abstractmethod
    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0,
                step_length:int = 25, move_time: int = 1500) -> bool:
        return False

    @abstractmethod
    def eyes(self, joint_id: int, pose_or_angle: Union[str, int], move_time: int = 1000) -> bool:
        return False

    @abstractmethod
    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2500) -> bool:
        return False

    @abstractmethod
    def arms(self, left_angle: int, right_angle: int, move_time: int) -> bool:
        return False

    @abstractmethod
    def celebrate(self, move_time: int = 4000) -> bool:
        return False

    @abstractmethod
    def circle_dance(self, side: str = 'right', move_time: int = 2500) -> bool:
        return False

    @abstractmethod
    def dance(self, side: str = 'right', move_time: int = 3000) -> bool:
        return False

    @abstractmethod
    def wiggle(self, move_time: int = 5000) -> bool:
        return False

    @abstractmethod
    def sidestep(self, side: str, steps: int = 1, step_length: int = 50,
            move_time: int = 1000) -> bool:
        return False

    @abstractmethod
    def set_volume(self, volume: int) -> bool:
        pass

    @abstractmethod
    def get_volume(self) -> int:
        return 0

    @abstractmethod
    def play_sound(self, name_or_freq_start: Union[str,float], 
            freq_end: Optional[float] = None, 
            duration: Optional[int] = None) -> bool:
        return False

    @abstractmethod
    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        return False

    @abstractmethod
    def write_gpio(self, gpio: int, value: int) -> bool:
        return False

    @abstractmethod
    def digitalread_gpio(self, gpio: int) -> bool:
        return False

    @abstractmethod
    def i2c_write(self, *byte_array: int) -> bool:
        return False

    @abstractmethod
    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        return False

    @abstractmethod
    def get_battery_voltage(self) -> float:
        return 0

    @abstractmethod
    def get_battery_remaining(self) -> float:
        return 0

    @abstractmethod
    def get_distance_sensor(self) -> Union[int, float]:
        return 0

    @abstractmethod
    def foot_on_ground(self, add_on_or_side: str) -> bool:
        return False

    @abstractmethod
    def foot_obstacle_sensed(self, add_on_or_side: str) -> bool:
        return False
        
    @abstractmethod
    def get_obstacle_sensor_reading(self, add_on_or_side: str) -> int:
        return 0

    @abstractmethod
    def get_ground_sensor_reading(self, add_on_or_side: str) -> int:
        return 0

    @abstractmethod
    def get_accelerometer(self, axis: Optional[str] = None, axisCode: int = 0) -> float:
        return 0

    @abstractmethod
    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        return False

    @abstractmethod
    def enable_safeties(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def fall_protection(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def motor_protection(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def battery_protection(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def buzz_prevention(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def lifelike_behaviour(self, enable: bool = True) -> bool:
        return False

    @abstractmethod
    def set_parameter(self, *byte_array: int) -> bool:
        return False

    @abstractmethod
    def save_calibration(self) -> bool:
        return False

    @abstractmethod
    def clear_calibration(self) -> bool:
        return False

    @abstractmethod
    def is_calibrated(self) -> bool:
        return False

    @abstractmethod
    def ros_command(self, *byte_array: int) -> bool:
        return False

    @abstractmethod
    def keyframe (self, time: float, num_of_msgs: int, msgs) -> List[bytes]:
        return False

    @abstractmethod
    def get_chatter(self) -> bytes:
        return False

    @abstractmethod
    def get_firmware_version(self) -> bool:
        return False

    @abstractmethod
    def _mute_serial(self) -> bool:
        return False

    @abstractmethod
    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        return False

    @abstractmethod
    def is_moving(self) -> bool:
        return False

    @abstractmethod
    def is_paused(self) -> bool:
        return False

    @abstractmethod
    def get_robot_status(self) -> Dict:
        return {}

    @abstractmethod
    def get_joints(self) -> Dict:
        return {}

    @abstractmethod
    def get_power_status(self) -> Dict:
        return {}

    @abstractmethod
    def get_add_ons_status(self) -> Dict:
        return {}

    @abstractmethod
    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        return {}

    @abstractmethod
    def add_on_query(self, add_on_name: str, data_to_write: bytes, num_bytes_to_read: int) -> Dict:
        return {}

    @abstractmethod
    def get_system_info(self) -> Dict:
        return {}

    @abstractmethod
    def set_marty_name(self, name: str) -> bool:
        return False

    @abstractmethod
    def get_marty_name(self) -> str:
        return ""

    @abstractmethod
    def is_marty_name_set(self) -> bool:
        return False

    @abstractmethod
    def get_hw_elems_list(self) -> List:
        return []

    @abstractmethod
    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        pass

    @abstractmethod
    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        return {}

    @abstractmethod
    def disco_off(self, add_on: str) -> bool :
        return False

    @abstractmethod
    def disco_pattern(self, pattern: int, add_on: str) -> bool :
        return False

    @abstractmethod   
    def disco_color(self, color: Union[str, Tuple[int, int, int]], add_on: str, region: Union[int, str]) -> bool:  
        return False

    @abstractmethod
    def disco_group_operation(self, disco_operation: Callable, whoami_type_codes: set, operation_kwargs: dict) -> bool:
        return False

    @abstractmethod
    def register_logging_callback(self, loggingCallback: Callable[[str],None]) -> None:
        pass

    @abstractmethod
    def register_publish_callback(self, messageCallback: Callable[[int],None]) -> None:
        pass

    @abstractmethod
    def register_report_callback(self, messageCallback: Callable[[str],None]) -> None:
        pass

    @abstractmethod
    def get_interface_stats(self) -> Dict:
        return {}

    @abstractmethod
    def preException(self, isFatal: bool) -> None:
        pass

    @abstractmethod
    def get_test_output(self) -> dict:
        return ""

    @abstractmethod
    def is_conn_ready(self) -> bool:
        return False

    @abstractmethod
    def send_file(self, filename: str, 
                progress_callback: Callable[[int, int], bool] = None,
                file_dest:str = "fs") -> bool:
        return False

    @abstractmethod
    def play_mp3(self, filename: str,
                progress_callback: Callable[[int, int], bool] = None) -> bool:
        return False

    @abstractmethod
    def get_file_list(self) -> List[str]:
        return []

    @abstractmethod
    def delete_file(self, filename: str) -> bool:
        return False


'''
Python library to communicate with Marty the Robot V1 and V2 by Robotical

Getting started:  
1) Import Marty from the martypy library  
2) Create a Marty object that connects the way you want  
3) Tell your Marty to dance  

```python
from martypy import Marty  
my_marty = Marty("wifi","192.168.0.53")  
my_marty.dance()
```

The tags :one: and :two: indicate when the method is available for Marty V1 :one: and Marty V2 :two:
'''
import time
from typing import Callable, Dict, List, Optional, Union, Tuple
from enum import Enum
from .ClientGeneric import ClientGeneric
from .ClientMV2 import ClientMV2
from .ClientMV1 import ClientMV1
from .Exceptions import (MartyCommandException,
                         MartyConfigException)
from .RICROSSerial import RICROSSerial

class Marty(object):

    CLIENT_TYPES = {
        'socket' : ClientMV1,
        'exp'    : ClientMV2,
        'usb'    : ClientMV2,
        'wifi'   : ClientMV2,
        'test'   : ClientMV2,
    }

    STOP_TYPE = {
        'clear queue'       : 0, # clear movement queue only (so finish the current movement)
        'clear and stop'    : 1, # clear movement queue and servo queues (freeze in-place)
        'clear and disable' : 2, # clear everything and disable motors
        'clear and zero'    : 3, # clear everything, and make robot return to zero
        'pause'             : 4, # pause, but keep servo and movequeue intact and motors enabled
        'pause and disable' : 5, # as 4, but disable motors too
    }

    JOINT_IDS = {
        'left hip'       : 0,
        'left twist'     : 1,
        'left knee'      : 2,
        'right hip'      : 3,
        'right twist'    : 4,
        'right knee'     : 5,
        'left arm'       : 6,
        'right arm'      : 7,
        'eyes'           : 8        
    }

    JOINT_STATUS = {
        'enabled'           : 0x01,
        'current limit now' : 0x02,
        'current limit long': 0x04,
        'busy'              : 0x08,
        'pos restricted'    : 0x10,
        'paused'            : 0x20,
        'comms ok'          : 0x80
    }

    ACCEL_AXES = {
        'x' : 0,
        'y' : 1,
        'z' : 2,
    }

    HW_ELEM_TYPES = [
        "SmartServo",
        "IMU",
        "I2SOut",
        "BusPixels",
        "GPIO",
        "FuelGauge",
        "PowerCtrl"
        ]

    HW_ELEM_IDS = {
        'LeftHip'        : 0,
        'LeftTwist'      : 1,
        'LeftKnee'       : 2,
        'RightHip'       : 3,
        'RightTwist'     : 4,
        'RightKnee'      : 5,
        'LeftArm'        : 6,
        'RightArm'       : 7,
        'Eyes'           : 8,
        'IMU0'           : 19,
    }

    ADD_ON_TYPE_NAMES = [
        "IRFoot"
    ]

    PUBLISH_TOPIC_SERVOS = RICROSSerial.ROSTOPIC_V2_SMART_SERVOS
    PUBLISH_TOPIC_ACCELEROMETER = RICROSSerial.ROSTOPIC_V2_ACCEL
    PUBLISH_TOPIC_POWER = RICROSSerial.ROSTOPIC_V2_POWER_STATUS
    PUBLISH_TOPIC_ADDONS = RICROSSerial.ROSTOPIC_V2_ADDONS
    PUBLISH_TOPIC_ROBOT_STATUS = RICROSSerial.ROSTOPIC_V2_ROBOT_STATUS

    class Disco(Enum):
        ARMS = {"00000088"}
        FEET = {"00000087"}
        EYES = {"00000089"}
        ALL = {"00000087", "00000088", "00000089"}

    def __init__(self,
                method: str,
                locator: str = "",
                extra_client_types: dict = dict(),
                blocking: Union[bool, None] = None,
                *args, **kwargs) -> None:
        '''
        Start a connection to Marty :one: :two:

        For example:

            * `Marty("wifi", "192.168.86.53")` to connect to Marty via WiFi on IP Address 192.168.0.53
            * `Marty("usb", "COM2")` on a Windows computer with Marty connected by USB cable to COM2
            * `Marty("usb", "/dev/tty.SLAB_USBtoUART")` on a Mac computer with Marty connected by USB cable to /dev/tty.SLAB_USBtoUART
            * `Marty("exp", "/dev/ttyAMA0")` on a Raspberry Pi computer with Marty connected by expansion cable to /dev/ttyAMA0

        **Blocking Mode**

        Each command that makes Marty move (e.g. `walk()`, `dance()`, `move_joint()`,
        but also `hold_position()`) comes with two modes - blocking and non-blocking.

        Issuing a command in blocking mode will make your program pause until Marty
        physically stops moving. Only then the next line of your code will be executed.

        In non-blocking mode, each movement command simply tells Marty what to do and
        returns immediately, meaning that your code will continue to execute while
        Marty is moving.

        Every movement command takes an optional `blocking` argument that can be used
        to choose the mode for that call. If you plan to use the same mode all or most
        of the time, it is better to to use the `Marty.set_blocking()` method or use
        the `blocking` constructor argument. The latter defaults to `True` (blocking)
        if not provided.

        Args:
            method: method of connecting to Marty - it may be: "usb",
                "wifi", "socket" (Marty V1) or "exp" (expansion port used to connect
                to a Raspberry Pi, etc)
            locator: location to connect to, depending on the method of connection this
                is the serial port name, network (IP) Address or network name (hostname) of Marty
                that the computer should use to communicate with Marty.
                If `method` is `"usb"` and there is only a single Marty connected,
                `locator` can be left out and the Marty will be found automatically.
                If multiple Martys are detected, one of them will be chosen
                arbitrarily and connected to.
            blocking: Default movement command mode for this `Marty` instance.
                * `True` (default): blocking mode
                * `False`: non-blocking mode

        Raises:
            * MartyConfigException if the parameters are invalid
            * MartyConnectException if Marty couldn't be contacted
        '''
        # Merge in any extra clients that have been added and check valid
        self.client: ClientGeneric = None
        self.CLIENT_TYPES = ClientGeneric.dict_merge(self.CLIENT_TYPES, extra_client_types)

        # Get and check connection parameters
        if type(method) is not str:
            raise MartyConfigException(f'Method must be one of {self.CLIENT_TYPES.keys()}')
        if '://' in method:
            method, _, locator = method.partition('://')
        if method.lower() not in self.CLIENT_TYPES.keys():
            raise MartyConfigException(f'Unrecognised method "{method}"')

        # Initialise the client class used to communicate with Marty
        client_cls = self.CLIENT_TYPES[method.lower()]
        self.client = client_cls(method.lower(), locator, *args, blocking=blocking, **kwargs)

        # Get Marty details
        self.client.start()

    def dance(self, side: str = 'right', move_time: int = 3000, blocking: Optional[bool] = None) -> bool:
        '''
        Boogie, Marty! :one: :two:
        Args:
            side: 'left' or 'right', which side to start on
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.dance(side, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def celebrate(self, move_time: int = 4000, blocking: Optional[bool] = None) -> bool:
        '''
        Coming soon! Same as `wiggle()` for now. :one: :two:
        Args:
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.celebrate(move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def wiggle(self, move_time: int = 4000, blocking: Optional[bool] = None) -> bool:
        '''
        Wiggle :two:
        Args:
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.wiggle(move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def circle_dance(self, side: str = 'right', move_time: int = 2500, blocking: Optional[bool] = None) -> bool:
        '''
        Circle Dance :two:
        Args:
            side: 'left' or 'right', which side to start on
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.circle_dance(side, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0,
                step_length:int = 25, move_time: int = 1500, blocking: Optional[bool] = None) -> bool:
        '''
        Make Marty walk :one: :two:
        Args:
            num_steps: how many steps to take
            start_foot: 'left', 'right' or 'auto', start walking with this foot Note: :two: Unless
                    you specify 'auto', all steps are taken with the same foot so
                    it only makes sense to use the start_foot argument with `num_steps=1`.
            turn: How much to turn (-100 to 100 in degrees), 0 is straight.
            step_length: How far to step (approximately in mm)
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.walk(num_steps, start_foot, turn, step_length, move_time)
        if result:
            num_steps = max(1, min(num_steps, 10))  # Clip num_steps
            self.client.wait_if_required(num_steps*move_time, blocking)
        return result
    
    def lift_foot(self, side: str) -> bool:
        '''
        Lift one of Marty's feet :two:
        Args:
            side: 'left' or 'right', which foot to lift
        Returns:
            True if Marty accepted the request
        '''
        return self.client.lift_foot(side)

    def lower_foot(self, side: str) -> bool:
        '''
        Lower one of Marty's feet :two:
        Args:
            side: 'left' or 'right', which foot to lower
        Returns:
            True if Marty accepted the request
        '''
        return self.client.lower_foot(side)
    
    def wave(self, side: str) -> bool:
        '''
        Wave :two:
        Args:
            None
        Returns:
            True if Marty accepted the request
        '''
        return self.client.wave(side)

    def get_ready(self, blocking: Optional[bool] = None) -> bool:
        '''
        Move Marty to the normal standing position and wiggle eyebrows :one: :two:
        Will also enable motors for Marty v1 :one:
        Args:
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.get_ready()
        if result:
            self.client.wait_if_required(4000, blocking)
        return result

    def stand_straight(self, move_time: int = 2000, blocking: Optional[bool] = None) -> bool:
        '''
        Move Marty to the normal standing position :one: :two:
        Args:
            move_time: How long (in milliseconds) Marty will take to reach the
                normal standing position. (Higher number means slower movement.)
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.stand_straight(move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def eyes(self, pose_or_angle: Union[str, int], move_time: int = 1000, blocking: Optional[bool] = None) -> bool:
        '''
        Move the eyes to a pose or an angle :one: :two:
        Args:
            pose_or_angle: 'angry', 'excited', 'normal', 'wide', or 'wiggle' :two: - alternatively
                           this can be an angle in degrees (which can be a negative number)
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.eyes(Marty.JOINT_IDS['eyes'], pose_or_angle, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2500, blocking: Optional[bool] = None) -> bool:
        '''
        Kick one of Marty's feet :one: :two:
        Args:
            side: 'left' or 'right', which foot to use
            twist: the amount of twisting do do while kicking (in degrees)
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.kick(side, twist, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def arms(self, left_angle: int, right_angle: int, move_time: int, blocking: Optional[bool] = None) -> bool:
        '''
        Move both of Marty's arms to angles you specify :one: :two:
        Args:
            left_angle: Angle of the left arm (degrees -100 to 100)
            right_angle: Position of the right arm (degrees -100 to 100)
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.arms(left_angle, right_angle, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def lean(self, direction: str, amount: Optional[int] = None, move_time: int = 1000,
             blocking: Optional[bool] = None) -> bool:
        '''
        Lean over in a direction :one: :two:
        Args:
            direction: `'left'`, `'right'`, `'forward'`, or `'back'`
            amount: How much to lean. The defaults and the exact meaning is
                different between Marty V1 and V2:
                - :one: If not specified or `None`, `amount` defaults to `50` (no
                        specific unit). Limit is -60/60 in either direction.
                - :two: If not specified or `None`, `amount` defaults to `29` degrees.
                        Limit for foward and back is 45 and limit for left and right is 60.
            move_time: How long this movement should last, in milliseconds.
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.lean(direction, amount, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def sidestep(self, side: str, steps: int = 1, step_length: int = 35,
            move_time: int = 1000, blocking: Optional[bool] = None) -> bool:
        '''
        Take sidesteps :one: :two:
        Args:
            side: 'left' or 'right', direction to step in
            steps: number of steps to take
            step_length: how broad the steps are (up to 127)
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.sidestep(side, steps, step_length, move_time)
        if result:
            steps = max(1, min(steps, 10))  # Clip steps
            self.client.wait_if_required(steps*move_time, blocking)
        return result

    def set_volume(self, volume: int) -> bool:
        '''
        Set the volume of Marty's sound :two:
        Args:
            volume: Volume to set (0-100)
        Returns:
            True if Marty accepted the request
        '''
        return self.client.set_volume(volume)

    def get_volume(self) -> int:
        '''
        Get the volume of Marty's sound :two:
        Args:
            None
        Returns:
            Volume of Marty's sound (0-100)
        '''
        return self.client.get_volume()

    def play_sound(self, name_or_freq_start: Union[str,int], 
            freq_end: Optional[int] = None, 
            duration: Optional[int] = None) -> bool:
        '''
        Play a named sound (Marty V2 :two:) or make a tone (Marty V1 :one:)
        Args:
            name_or_freq_start: name of the sound, e.g. 'excited' or 'no_way' :two:
            name_or_freq_start: starting frequency, min 20, max 20000, Hz :one:
            freq_end:  ending frequency, min 20, max 20000, Hz :one:
            duration: milliseconds, maximum 5000 :one:
        Returns:
            True if Marty accepted the request
        '''
        return self.client.play_sound(name_or_freq_start, freq_end, duration)

    def get_accelerometer(self, axis: Optional[str] = None) -> float:
        '''
        Get the latest value from the Marty's accelerometer :one: :two:
        Args:
            axis: (optional) 'x', 'y' or 'z' OR no parameter at all (see returns below)
        Returns:  
            * The acceleration value from the axis (if axis specified)  
            * A tuple containing x, y and z values (if no axis) :two:
            Note that the returned value will be 0 if no value is available
        Raises:
            MartyCommandException if the axis is unknown
        '''
        axisCode = 0
        if (axis is not None) and (type(axis) is str):
            if axis not in self.ACCEL_AXES:
                self.client.preException(True)
                raise MartyCommandException("Axis must be one of {}, not '{}'"
                                            "".format(set(self.ACCEL_AXES.keys()), axis))
            axisCode = self.ACCEL_AXES.get(axis, 0)
        return self.client.get_accelerometer(axis, axisCode)

    def is_moving(self) -> bool:
        '''
        Check if Marty is moving :two:
        Args:
            none
        Returns:
            True if Marty is moving
        '''
        return self.client.is_moving()

    def stop(self, stop_type: Optional[str] = None) -> bool:
        '''
        Stop Marty's movement  :one: :two:  

        You can also control what way to "stop" you want with the parameter stop_type. For instance:  
        
        * 'clear queue' to finish the current movement before stopping (clear any queued movements)
        * 'clear and stop' stop immediately (and clear queues)
        * 'clear and disable' :one: stop and disable the robot
        * 'clear and zero' stop and move back to get_ready pose
        * 'pause' pause motion
        * 'pause and disable' :one: pause motion and disable the robot

        Args:
            stop_type: the way to stop - see the options above

        Raises:
            MartyCommandException if the stop_type is unknown
        '''
        # Default to plain "stop"
        stopCode = 1 
        if stop_type is not None:
            if stop_type not in self.STOP_TYPE:
                self.client.preException(True)
                raise MartyCommandException("Unknown stop_type '{}', not in Marty.STOP_TYPE"
                                            "".format(stop_type))
            stopCode = self.STOP_TYPE.get(stop_type, stopCode)
        return self.client.stop(stop_type, stopCode)

    def resume(self) -> bool:
        '''
        Resume Marty's movement after a pause :two:
        Returns:
            True if Marty accepted the request
        '''
        return self.client.resume()

    def hold_position(self, hold_time: int, blocking: Optional[bool] = None) -> bool:
        '''
        Hold Marty at its current position :two:
        Args:
            hold_time, time to hold position in milli-seconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
                Holding position counts as movement because Marty is using its motors to
                actively resist any attempts to move its joints.
        Returns:
            True if Marty accepted the request
        '''
        # Default to plain "stop"
        result = self.client.hold_position(hold_time)
        if result:
            self.client.wait_if_required(hold_time, blocking)
        return result

    def is_paused(self) -> bool:
        '''
        Check if Marty is paused :two:
        Returns:
            True if Marty is paused
        '''
        return self.client.is_paused()

    def is_blocking(self) -> bool:
        '''
        Check the default movement command behaviour of this Marty.  :one: :two:
        Returns:
            `True` if movement commands block by default
        '''
        return self.client.is_blocking()

    def set_blocking(self, blocking: bool):
        '''
        Change whether movement commands default to blocking or non-blocking behaviour
        for this Marty.  :one: :two:

        The blocking behaviour can also be specified on a per-command basis using the
        `blocking=` argument which takes precedence over Marty's overall setting.

        Args:
            blocking: whether or not to block by default
        '''
        self.client.set_blocking(blocking)

    def move_joint(self, joint_name_or_num: Union[int, str], position: int, move_time: int,
                   blocking: Optional[bool] = None) -> bool:
        '''
        Move a specific joint to a position :one: :two:
        Args:
            joint_name_or_num: joint to move, see the Marty.JOINT_IDS dictionary (can be name or number)
            position: angle in degrees
            move_time: how long this movement should last, in milliseconds
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        Returns:
            True if Marty accepted the request
        Raises:
            MartyCommandException if the joint_name_or_num is unknown
        '''
        jointIDNo = joint_name_or_num
        if type(joint_name_or_num) is str:
            if joint_name_or_num not in self.JOINT_IDS:
                self.client.preException(True)
                raise MartyCommandException("Joint must be one of {}, not '{}'"
                                            "".format(set(self.JOINT_IDS.keys()), joint_name_or_num))
            jointIDNo = self.JOINT_IDS.get(joint_name_or_num, 0)
        result = self.client.move_joint(jointIDNo, position, move_time)
        if result:
            self.client.wait_if_required(move_time, blocking)
        return result

    def get_joint_position(self, joint_name_or_num: Union[int, str]) -> float:
        '''
        Get the position (angle in degrees) of a joint :two:
        Args:
            joint_name_or_num: see the Marty.JOINT_IDS dictionary (can be name or number)
        Returns:
            Angle of the joint in degrees
        Raises:
            MartyCommandException if the joint_name_or_num is unknown            
        '''
        jointIDNo = joint_name_or_num
        if type(joint_name_or_num) is str:
            if joint_name_or_num not in self.JOINT_IDS:
                self.client.preException(True)
                raise MartyCommandException("Joint must be one of {}, not '{}'"
                                            "".format(set(self.JOINT_IDS.keys()), joint_name_or_num))
            jointIDNo = self.JOINT_IDS.get(joint_name_or_num, 0)
        return self.client.get_joint_position(jointIDNo)

    def get_joint_current(self, joint_name_or_num: Union[int, str]) -> float:
        '''
        Get the current (in milli-Amps) of a joint :one: :two:
        This can be useful in detecting when the joint is working hard and is related
        to the force which the joint's motor is exerting to stay where it is
        Args:
            joint_name_or_num: see the Marty.JOINT_IDS dictionary (can be name or number)
        Returns:
            the current of the joint in milli-Amps (this will be 0 if the joint current is unknown)
        Raises:
            MartyCommandException if the joint_name_or_num is unknown
        '''
        jointIDNo = joint_name_or_num
        if type(joint_name_or_num) is str:
            if joint_name_or_num not in self.JOINT_IDS:
                self.client.preException(True)
                raise MartyCommandException("Joint must be one of {}, not '{}'"
                                            "".format(set(self.JOINT_IDS.keys()), joint_name_or_num))
            jointIDNo = self.JOINT_IDS.get(joint_name_or_num, 0)
        return self.client.get_joint_current(jointIDNo)

    def get_joint_status(self, joint_name_or_num: Union[int, str]) -> int:
        '''
        Get information about a joint :two:
        This can be helpful to find out if the joint is working correctly and if it is
        moving at the moment, etc
        Args:
            joint_name_or_num: see the Marty.JOINT_IDS dictionary (can be name or number)
        Returns:
            a code number which is the sum of codes in the Marty.JOINT_STATUS dictionary
            will be 0 if the joint status is unknown
        Raises:
            MartyCommandException if the joint_name_or_num is unknown
        '''
        jointIDNo = joint_name_or_num
        if type(joint_name_or_num) is str:
            if joint_name_or_num not in self.JOINT_IDS:
                self.client.preException(True)
                raise MartyCommandException("Joint must be one of {}, not '{}'"
                                            "".format(set(self.JOINT_IDS.keys()), joint_name_or_num))
            jointIDNo = self.JOINT_IDS.get(joint_name_or_num, 0)
        return self.client.get_joint_status(jointIDNo)

    def get_distance_sensor(self) -> Union[int, float]:
        '''
        Get the latest value from the distance sensor :one: :two:
        Returns:
            The distance sensor reading. The meaning of the returned value is different 
            between Marty V1 and V2:
                - :one: Returns a raw distance sensor reading as a `float`.
                - :two: Returns the distance in millimeters as `int`.
            Both will return 0 if no distance sensor is found.
        '''
        return self.client.get_distance_sensor()

    def speak(self, words: str = "hello", voice: str = "alto", blocking: Optional[bool] = False) -> bool:
        '''
        Make Marty speak :two:
        Args:
            words: what to say
            voice: 'alto', 'tenor' or 'chipmunk' :two:
            blocking: whether to wait for speech to finish before returning
        Returns:
            True if Marty accepted the request
        '''
        result = self.client.speak(words, voice)
        if result:
            if blocking:
                self.client.wait_if_required(5000, blocking)
        return result

    def foot_on_ground(self, add_on_or_side: str) -> bool:
        '''
        Checks whether the foot is on a surface :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor, name of an IR sensor, `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            A boolean for whether the addon detects the ground. `True` for ground detected, `False` otherwise.
        '''
        return self.client.foot_on_ground(add_on_or_side)

    def foot_obstacle_sensed(self, add_on_or_side: str) -> bool:
        '''
        Checks whether there is an obstacle in front of the foot :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor, name of an IR sensor, `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            A boolean for whether the addon detects and obstacle. `True` for obstacle detected, `False` otherwise.
        '''
        return self.client.foot_obstacle_sensed(add_on_or_side)

    def get_obstacle_sensor_reading(self, add_on_or_side: str) -> int:
        '''
        Gets a raw obstacle sensor reading from an IR or color sensor :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor, name of an IR sensor, `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            Raw reading of obstacle sensor data from the add on.
        '''
        return self.client.get_obstacle_sensor_reading(add_on_or_side)
    
    def get_color_sensor_color(self, add_on_or_side: str) -> int:
        '''
        Gets the colour detected by a colour sensor :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            one of  "yellow", "green", "blue", "purple", "red", "air", "unclear"
        '''
        return self.client.get_color_sensor_color(add_on_or_side)
    
    def get_color_sensor_hex(self, add_on_or_side: str) -> str:
        '''
        Gets the colour detected by a colour sensor :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            The hex code of the colour detected by the sensor
        '''
        return self.client.get_color_sensor_hex(add_on_or_side)
    
    def get_color_sensor_value_by_channel(self, add_on_or_side: str, channel: str) -> int:
        '''
        Gets the value of a colour sensor channel :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
            channel: Takes in the name of a channel `'clear'` `'red'`, `'green'`, or `'blue'`
        Returns:
            The value of the channel
        '''
        return self.client.get_color_sensor_value_by_channel(add_on_or_side, channel)

    def get_ground_sensor_reading(self, add_on_or_side: str) -> int:
        '''Gets a raw ground sensor reading from an IR or color sensor :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor, name of an IR sensor, `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            Raw reading of ground sensor data from the add on.
        '''
        return self.client.get_ground_sensor_reading(add_on_or_side)

    def get_battery_remaining(self) -> float:
        '''
        Get the battery remaining percentage :two:
        Returns:
            The battery remaining capacity in percent
        '''
        return self.client.get_battery_remaining()

    def save_calibration(self) -> bool:
        '''
        Set the current motor positions as the zero positions :one: :two:

        BE CAREFUL, this can cause unexpected movement or self-interference
        '''
        return self.client.save_calibration()

    def clear_calibration(self) -> bool:
        '''
        Mark the current calibration as invalid :one: :two:

        This has no immediate physical effect. Marty will still remember the last
        calibration but will report that it needs to be calibrated again. (You may
        notice that a "Calibrate" button appears in the app for example.)
        '''
        return self.client.clear_calibration()

    def is_calibrated(self) -> bool:
        '''
        Check if Marty is calibrated :two:
        '''
        return self.client.is_calibrated()

    def get_robot_status(self) -> Dict:
        '''
        Get status of Marty the Robot :two:
        Args:
            none
        Returns:
            Dictionary containing:
                "workQCount" number of work items (movements) that are queued up
                "isMoving": True if Marty is moving
                "isPaused": True if Marty is paused
                "isFwUpdating": True if Marty is doing an update
        '''
        return self.client.get_robot_status()

    def get_joints(self) -> Dict: 
        '''
        Get information on all of Marty's joints :two:
        Args:
            none
        Returns:
            Dictionary containing dictionaries (one for each joint) each of which contain:
                "IDNo": the joint identification number (see Marty.JOINT_IDS)
                "name": the name of the joint
                "pos": the angle of the joint
                "current": the joint current (in milli-Amps)
                "enabled": True if the servo is enabled
                "commsOK": True if the servo is communicating ok
                "flags": joint status flags (see Marty.JOINT_STATUS)
        '''
        return self.client.get_joints()

    def get_power_status(self) -> Dict:
        '''
        Get information on Marty's battery and power supply :two:
        Args:
            none
        Returns:
            Dictionary containing:
                "battRemainCapacityPercent": remaining battery capacity in percent
                "battTempDegC": battery temperature in degrees C
                "battRemainCapacityMAH": remaining battery capacity in milli-Amp-Hours
                "battFullCapacityMAH": capacity of the battery when full in milli-Amp-Hours
                "battCurrentMA": current the battery is supplying (or being charged with) milli-Amps
                "power5VOnTimeSecs": number of seconds the power to joints and add-ons has been on
                "powerUSBIsConnected": True if USB is connected
                "power5VIsOn": True if power to the joints and add-ons is turned on

                 Other values for internal use

            **Note:** Some keys may not be included if Marty reports that the
                      corresponding information is not available.
        '''
        return self.client.get_power_status()
    
    def get_add_ons_status(self) -> Dict:
        '''
        Get latest information for all add-ons :two:
        Args:
            none
        Returns:
            Dictionary containing dictionaries (one for each add-on) each of which contain:
                "IDNo": the add-on identification number
                "name": the name of the add-on
                "type": the type of the add-on (see Marty.ADD_ON_TYPE_NAMES but it may not be in this list)
                "whoAmITypeCode": a code which can be used for further add-on identification
                "valid": True if the data is valid
                "data": 10 bytes of data from the add-on - the format of this data depends
                        on the type of add-on
        '''
        return self.client.get_add_ons_status()

    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        '''
        Get latest information for a single add-on :two:
        Args:
            add_on_name_or_id: either the name or the id (number) of an add-on
        Returns:
            Dictionary containing:
                "IDNo": the add-on identification number
                "valid": True if the data is valid
                "data": 10 bytes of data from the add-on - the format of this data depends
                        on the type of add-on
        '''
        return self.client.get_add_on_status(add_on_name_or_id)

    def add_on_query(self, add_on_name: str, data_to_write: bytes, num_bytes_to_read: int) -> Dict:
        '''
        Write and read an add-on directly (raw-mode) :two:
        Args:
            add_on_name: name of the add-on (see get_add_ons_status() at the top level or response to
                `addon/list` REST API command)
            data_to_write: can be zero length if nothing is to be written, the first byte will generally
                be the register or opcode of the add-on
            num_bytes_to_read: number of bytes to read from the device - can be zero
        Returns:
            Dict with keys including:
                "rslt" - the result which will be "ok" if the query succeeded
                "dataRead" - the data read from the add-on
        '''
        return self.client.add_on_query(add_on_name, data_to_write, num_bytes_to_read)

    def get_system_info(self) -> Dict:
        '''
        Get information about Marty :two:
        Args:
            none
        Returns:
            Dictionary containing:
                "HardwareVersion": string containing the version of Marty hardware
                                "1.0" for Marty V1
                                "2.0" for Marty V2
                                other values for later versions of Marty
                "SystemName": the name of the physical hardware in Marty - this will be 
                              RicFirmwareESP32 for Marty V2 and
                              MartyV1 for Marty V1
                "SystemVersion": a string in semantic versioning format with the version
                                 of Marty firmware (e.g. "1.2.3")
                "SerialNo": serial number of this Marty
                "MAC": the base MAC address of the Marty
                "RicHwRevNo": the revision number of the RIC hardware
        '''
        return self.client.get_system_info()

    def set_marty_name(self, name: str) -> bool:
        '''
        Set Marty's name :two:
        Args:
            name to call Marty
        Returns:
            True if successful in setting the name
        '''
        return self.client.set_marty_name(name)

    def get_marty_name(self) -> str:
        '''
        Get Marty's name :two:
        Args:
            none
        Returns:
            the name given to Marty
        '''
        return self.client.get_marty_name()

    def is_marty_name_set(self) -> bool:
        '''
        Check if Marty's name is set :two:
        Args:
            none
        Returns:
            True if Marty's name is set
        '''
        return self.client.is_marty_name_set()

    def get_hw_elems_list(self) -> List:
        '''
        Get a list of all of the hardware elements on Marty :two:
        Args:
            none
        Returns:
            List containing a dictionary for each hardware element, each element is in the form:
                "name": name of the hardware element
                "type": type of element, see Marty.HW_ELEM_TYPES, other types may appear as add-ons
                "busName": name of the bus the element is connected to
                "addr": address of the element if it is connected to a bus
                "addrValid": 1 if the address is valid, else 0
                "IDNo": identification number of the element
                "whoAmI": string from the hardware which may contain additional identification
                "whoAmITypeCode": string indicating the type of hardware
                "SN": serial number of the hardware element
                "versionStr": version string of hardware element in semantic versioning (semver) format
                "commsOK": 1 if the element is communicating ok, 0 if not
        '''
        return self.client.get_hw_elems_list()

    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        '''
        Send a command in RIC REST format to Marty :two:  

        This is a special purpose command which you can use to do advanced
        control of Marty
        Args:
            ricRestCmd: string containing the command to send to Marty
        Returns:
            None
        '''
        self.client.send_ric_rest_cmd(ricRestCmd)

    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        '''
        Send a command in RIC REST format to Marty and wait for reply :two:  

        This is a special purpose command which you can use to do advanced
        control of Marty
        Args:
            ricRestCmd: string containing the command to send to Marty
        Returns:
            Dictionary containing the response received from Marty
        '''
        return self.client.send_ric_rest_cmd_sync(ricRestCmd)

    def get_motor_current(self, motor_id: int) -> float:
        '''
        Get current flowing through a joint motor :one: :two:
        Args:
            motor_id, integer >= 0 (non-negative) selects which motor to query
        Returns:
            Instantaneous current sense reading from motor `motor_id`
        '''
        return self.get_joint_current(motor_id)

    def is_conn_ready(self) -> bool:
        '''
        Check if the robot is connected and the connection is ready to accept
        commands :two:
        Args:
            None
        Returns:
            True if the robot is connected and ready
        '''
        return self.client.is_conn_ready()
        
    def disco_off(self, add_on: Union[Disco, str] = Disco.ALL, api = 'led') -> bool:
        '''
        Turn disco add on LEDs off :two:
        Args:
            add_on: add on name of which the function applies to
            api: which API to use, 'raw_query' or 'led'
        Returns:
            True if Marty accepted the request
        '''
        if type(add_on) is str:
            return self.client.disco_off(add_on, api)
        else:
            return self.client.disco_group_operation(self.client.disco_off, add_on.value, {'api':api}, api=api)

    def disco_pattern(self, pattern: int, add_on: Union[Disco, str] = Disco.ALL) -> bool:
        '''
        DEPRECATED: use disco_named_pattern instead
        Turn on a pattern of lights on the disco LED add on :two:
        Args:
            pattern: 1 or 2, pattern of lights that user wants to use
            add_on: add on name of which the function applies to
        Returns:
            True if Marty accepted the request
        '''
        if type(add_on) is str:
            return self.client.disco_pattern(pattern, add_on)
        else:
            return self.client.disco_group_operation(self.client.disco_pattern, add_on.value, {'pattern':pattern}, api='raw_query')

    def disco_named_pattern(self, add_on: str, pattern: str) -> bool:
        '''
        Turn on a named pattern of lights on the disco LED add on :two:
        Args:
            add_on: add on name of which the function applies to
            pattern: name of the pattern to use
        Returns:
            True if Marty accepted the request
        '''
        if type(add_on) is str:
            return self.client.disco_named_pattern(add_on, pattern)
        else:
            return self.client.disco_group_operation(self.client.disco_named_pattern, add_on.value, {'pattern':pattern}, api='led')

    def disco_color(self, color: Union[str, Tuple[int, int, int]] = 'white', 
                    add_on: Union[Disco, str] = Disco.ALL, 
                    region: Union[int, str] = 'all',
                    api = 'led'
                    ) -> bool:
        '''
        Turn on disco add on LED lights to a specific color :two:
        Args:
            color: color to switch the LEDs to; takes in a hex code, RGB tuple of integers between 0-255, 
                   or one of the built in colors: white, red, blue, yellow, green, teal, pink, purple, orange
            add_on: add on name of which the function applies to
            region: 0,1,2; region on the add on
        Returns:
            True if Marty accepted the request
        '''
        if type(add_on) is str:
            if api == 'raw_query':
                return self.client.disco_color(color, add_on, region)
            elif api == 'led':
                return self.client.disco_color_led_api(color, add_on, region)
        else:
            if api == 'raw_query':
                return self.client.disco_group_operation(self.client.disco_color, add_on.value, {'color':color, 'region':region}, api='raw_query')
            elif api == 'led':
                return self.client.disco_group_operation(self.client.disco_color_led_api, add_on.value, {'color':color, 'region':region}, api='led')
                # raise MartyCommandException("Disco LED API not supported for disco group operation. Please use addon name")
    
    def disco_color_specific_led(self, color: Union[str, Tuple[int, int, int]], add_on: str, add_on_who_am_i: str, led_id: int) -> bool:
        '''
        Turn on disco add on specific LED light to a color :two:
        Args:
            color: color to switch the LEDs to; takes in a hex code, RGB tuple of integers between 0-255, 
                   or one of the built in colors: white, red, blue, yellow, green, teal, pink, purple, orange
            add_on: add on name of which the function applies to
            add_on_who_am_i: who am i string of the add on
            led_id: ID of the LED to change
        Returns:
            True if Marty accepted the request
        '''
        return self.client.disco_color_specific_led(color, add_on, add_on_who_am_i, led_id)

    def disco_color_eyepicker(self, colours: Union[str, List[str]], add_on: str) -> bool:
        '''
        Turn on disco add on specific LED lights to specific colors :two:
        Args:
            colours: list of colors to switch the LEDs to; takes in a hex code, the position in the list corresponds to the LED ID
            add_on: add on name of which the function applies to
        Returns:
            True if Marty accepted the request
        '''
        return self.client.disco_color_eyepicker(colours, add_on)

    def rgb_operator(self, r: int, g: int, b: int) -> list[int]:
        '''
        Takes in RGB values and returns them as tuples :one: :two:
        Args:
            r: red value
            g: green value
            b: blue value
        Returns:
            List of RGB values after the operation
        '''
        return (r, g, b)

    def hsv_operator(self, h: int, s: int, v: int) -> list[int]:
        '''
        Takes in HSL values and returns them in RGB tuples :one: :two:
        Args:
            h: hue value
            s: saturation value
            l: lightness value
        Returns:
            List of RGB values after the operation
        '''
        # first make sure s and v are 0-1
        if s > 1:
            s = s / 100
        if v > 1:
            v = v / 100
        h = h % 360
        if h < 0:
            h += 360
        s = max(0, min(s, 1))
        v = max(0, min(v, 1))

        i = int(h // 60)
        f = (h / 60) - i
        p = v * (1 - s)
        q = v * (1 - (s * f))
        t = v * (1 - (s * (1 - f)))

        if i == 0:
            r, g, b = v, t, p
        elif i == 1:
            r, g, b = q, v, p
        elif i == 2:
            r, g, b = p, v, t
        elif i == 3:
            r, g, b = p, q, v
        elif i == 4:
            r, g, b = t, p, v
        elif i == 5:
            r, g, b = v, p, q
        else:
            r, g, b = 0, 0, 0  # Fallback, though this shouldn't happen

        return [int(r * 255), int(g * 255), int(b * 255)]

    def function_led(self, colour: Tuple[int, int, int], breathe: str = "on", breath_ms: int = 1000) -> bool:
        '''
        Turn on function LED light to a color :two:
        Args:
            colour: color to switch the LEDs to; takes in an RGB tuple of integers between 0-255
            breathe: "on"/"breath"
            breath_ms: time in milliseconds for the LED to breath
        Returns:
            True if Marty accepted the request
        '''
        return self.client.function_led(colour, breathe, breath_ms)
    
    def function_led_off(self) -> bool:
        '''
        Turn off function LED light :two:
        Args:
            none
        Returns:
            True if Marty accepted the request
        '''
        return self.client.function_led_off()

    ''' 
    ============================================================
    The following commands are for Marty V1 Only
    ============================================================
    '''

    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        '''
        Toggle power to motors :one:
        Args:
            enable: True/False toggle
            clear_queue: Default True, prevents unfinished but 'muted' motions
                         from jumping as soon as motors are enabled
        '''
        return self.client.enable_motors(enable, clear_queue)

    def enable_safeties(self, enable: bool = True) -> bool:
        '''
        Tell the board to turn on 'normal' safeties :one:
        '''
        return self.client.enable_safeties(enable)

    def fall_protection(self, enable: bool = True) -> bool:
        '''
        Toggle fall protections :one:
        Args:
            enable: True/False toggle
        '''
        return self.client.fall_protection(enable)

    def motor_protection(self, enable: bool = True) -> bool:
        '''
        Toggle motor current protections :one:
        Args:
            enable: True/False toggle
        '''
        return self.client.motor_protection(enable)

    def battery_protection(self, enable: bool = True) -> bool:
        '''
        Toggle low battery protections :one:
        Args:
            enable: True/False toggle
        '''
        return self.client.battery_protection(enable)

    def buzz_prevention(self, enable: bool = True) -> bool:
        '''
        Toggle motor buzz prevention :one:
        Args:
            enable: True/False toggle
        '''
        return self.client.buzz_prevention(enable)

    def lifelike_behaviour(self, enable: bool = True) -> bool:
        '''
        Tell the robot whether it can or can't move now and then in a lifelike way when idle. :one:
        Args:
            enable: True/False toggle
        '''
        return self.client.lifelike_behaviour(enable)

    def ros_command(self, *byte_array: int) -> bool:
        '''
        Low level proxied access to the ROS Serial API between 
        the modem and main controller :one:
        '''
        return self.client.ros_command(*byte_array)

    def keyframe (self, time: float, num_of_msgs: int, msgs) -> List[bytes]:
        '''
        Takes in information about movements and generates keyframes
        returns a list of bytes :one:
        Args:
            time: time (in seconds) taken to complete movement
            num_of_msgs: number of commands sent
            msgs: commands sent in the following format [(ID CMD), (ID CMD), etc...]
        '''
        return self.client.keyframe(time, num_of_msgs, msgs)

    def get_chatter(self) -> bytes:
        '''
        Return chatter topic data (variable length) :one:
        '''
        return self.client.get_chatter()

    def get_firmware_version(self) -> bool:
        '''
        Ask the board to print the firmware version over chatter :one:
        '''
        return self.client.get_firmware_version()

    def _mute_serial(self) -> bool:
        '''
        Mutes the internal serial line on RIC. Depends on platform and API :one:
        NOTE: Once you've done this, the Robot will ignore you until you cycle power.
        '''
        return self.client.mute_serial()

    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        '''
        Formats message into ROS serial format and
        returns formatted message as a list :one:  

        Calls ros_command with the processed message if send is True.
        More information about the ROS serial format can be
        found here: http://wiki.ros.org/rosserial/Overview/Protocol
        '''
        return self.client.ros_serial_formatter(topicID, send, message)

    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        '''
        Configure a GPIO pin :one:
        Args:
            gpio: pin number between 0 and 7
            mode: choose from: 'digital in','analog in' or 'digital out'
        '''
        return self.client.pinmode_gpio(gpio, mode)

    def write_gpio(self, gpio:int, value: int) -> bool:
        '''
        Write a value to a GPIO port :one:
        '''
        return self.client.write_gpio(gpio, value)

    def digitalread_gpio(self, gpio: int) -> bool:
        '''
        Read from GPIO :one:
        Args:
            GPIO pin number, >= 0 (non-negative)
        Returns:
            Returns High/Low state of a GPIO pin
        '''
        return self.client.digitalread_gpio(gpio)

    def set_parameter(self, *byte_array: int) -> bool:
        '''
        Set board parameters :one:
        Args:
            byte_array: a list in the following format [paramID, params]
        '''
        return self.client.set_parameter(byte_array)

    def i2c_write(self, *byte_array: int) -> bool:
        '''
        Write a bytestream to the i2c port. :one:  

        The first byte should be the address, following from that
        the datagram folows standard i2c spec
        '''
        return self.client.i2c_write(*byte_array)

    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        '''
        Write a formatted bytestream to the i2c port. :one:  

        The bytestream is formatted in the ROS serial format.

        address: the other device's address
        '''
        return self.client.i2c_write_to_ric(address, byte_array)

    def i2c_write_to_rick(self, address: int, byte_array: bytes) -> bool:
        '''
        Write a formatted bytestream to the i2c port. :one:  

        The bytestream is formatted in the ROS serial format.
        address: the other device's address
        '''
        return self.client.i2c_write_to_ric(address, byte_array)

    def get_battery_voltage(self) -> float:
        '''
        Get the voltage of the battery :one:
        Returns:
            The battery voltage reading as a float in Volts
        '''
        return self.client.get_battery_voltage()

    def hello(self, blocking: Optional[bool] = None) -> bool:
        '''
        Zero joints and wiggle eyebrows :one:
        Args:
            blocking: Blocking mode override; whether to wait for physical movement to
                finish before returning. Defaults to the value returned by `self.is_blocking()`.
        '''
        result = self.client.hello()
        if result:
            self.client.wait_if_required(4000, blocking)
        return result

    def discover(self) -> List[str]:
        '''
        Try and find us some Martys! :one:
        '''
        return self.client.discover()

    ''' 
    ============================================================
    Helper commands that you probably won't need to use directly
    ============================================================
    '''

    def __del__(self) -> None:
        '''
        Marty is stopping
        '''
        self.close()

    def close(self) -> None:
        '''
        Close connection to Marty
        '''
        if self.client:
            self.client.close()

    def register_logging_callback(self, loggingCallback: Callable[[str], None]) -> None:
        '''
        Register a callback function to be called on every log message from RIC. :two:

        Log messages are used mainly for debugging and error reporting.
        Args:
            messageCallback: a callback function (with one string argument - the log message)
        Returns:
            None
        '''
        if self.client:
            self.client.register_logging_callback(loggingCallback)

    def register_publish_callback(self, messageCallback: Callable[[int],None]) -> None:
        '''
        Register a callback function to be called on every message published by RIC. :two:

        RIC publishes information like the accelerometer and joint positions constantly.
        If registered, the callback is called after the message is fully decoded, so a common
        use-case is to check the topic (the int passed to the callback) to see if the information
        is of interest (for example topic == Marty.PUBLISH_TOPIC_ACCELEROMETER if new accelerometer 
        data is available). Then get the changed information using the regular get_accelerometer method
        (or get_joints, etc for other data).
        Args:
            messageCallback: a callback function (with one argument - the topic code of the published 
                    message) that will be called on every message published by RIC.
        Returns:
            None
        '''
        if self.client:
            self.client.register_publish_callback(messageCallback)

    def register_report_callback(self, messageCallback: Callable[[str],None]) -> None:
        '''
        Register a callback function to be called on every report message recieved from RIC. :two:

        Report messages are used for alert conditions such as Falling, Over Current, etc.
        The report message (passed in the callback function str) is a JSON string.
        The elements of the JSON string include:
            msgType: the type of report message (this will be "warn" for all alerts)
            msgBody: the message text (this is "freeFallDet" or "overCurrentDet")
            IDNo: the IDNo of the element that generated the report (the codes for these
                are in the Marty.HW_ELEM_IDS dictionary but additional ones may appear if
                add-ons create warnings - the IDNos of add-ons are not fixed numbers but you
                can see the IDNo of add-ons by using the get_add_ons_status() method)
        Args:
            messageCallback: a callback function (with one string argument) 
                    that will be called on every message published by RIC.
        Returns:
            None
        '''
        if self.client:
            self.client.register_report_callback(messageCallback)

    def get_interface_stats(self) -> Dict:
        '''
        Get interface statistics from Marty. :two:

        This is mainly useful for debugging and troubleshooting of martypy itself.
        Returns:
            A dictionary of interface statistics including:
                roundTripAvgMS: average round trip time for command/response messages in milliseconds
                msgRxRatePS: max number of messages received per second
                msgTxRatePS: max number of messages sent per second
                unmatched: number of unmatched messages
                matched: number of matched messages
                unnumbered: number of unnumbered messages
                timedOut: number of messages that timed out
                uploadBPS: upload speed in bytes per second (for file uploads, etc)
                rxCount: number of messages received
                txCount: number of messages sent
            The dictionary may also include several records of the form:
                <topic>PS: number of messages published on each topic where <topic> is the topic name
                       and can be servos, imu, robot, power or addons
        '''
        return self.client.get_interface_stats()

    def get_test_output(self) -> str:
        '''
        This is only used for self-testing of martypy when the test interface is chosen instead of
        wifi, serial, etc :two:
        '''
        return self.client.get_test_output()

    def send_file(self, filename: str, 
                progress_callback: Callable[[int, int], bool] = None,
                file_dest:str = "fs") -> bool:
        '''
        Send a file to Marty. :two:

        Args:
            filename: the name of the file to send
            progress_callback: callback used to indicate how file send is progressing, callable takes three params
                    which are bytesSent and totalBytes and
                    returns a bool which should be True to continue the file upload or False to abort
            file_dest: "fs" to upload to file system, "ricfw" for new RIC firmware
        Returns:
            True if the file was sent successfully
        Throws:
            OSError: operating system exceptions
            May throw other exceptions so include a general exception handler
        '''
        return self.client.send_file(filename, progress_callback, file_dest)

    def play_mp3(self, filename: str,
                progress_callback: Callable[[int, int], bool] = None) -> bool:
        '''
        Play an mp3 file on the robot. :two:
        Args:
            filename: the name of the mp3 file to play
        Returns:
            True if the file was played successfully
        '''
        return self.client.play_mp3(filename, progress_callback)

    def get_file_list(self) -> List[str]:
        '''
        Get a list of files on the robot. :two:
        Args:
            None
        Returns:
            A list of files on the robot
        '''
        return self.client.get_file_list()

    def delete_file(self, filename: str) -> bool:
        '''
        Delete a file on the robot. :two:
        Args:
            filename: the name of the file to delete
        Returns:
            True if the file was deleted successfully
        '''
        return self.client.delete_file(filename)



import requests
import json
import time
from urllib.parse import quote
# from pydub import AudioSegment
# from pydub.playback import play
from io import BytesIO
# from .Exceptions import (MartyConnectException,
#                          MartyCommandException)
import io


# Voice IDs
VOICES = {
    "ALTO_ID": "ALTO",
    "KITTEN_ID": "KITTEN",
    "GIANT_ID": "GIANT",
    "TENOR_ID": "TENOR",
    "ALIEN_ID": "ALIEN",
    "THUNDER_ID": "THUNDER",
    "STONE_ID": "STONE",
    "RUMBLE_ID": "RUMBLE",
    "ECHO_ID": "ECHO",
    "DRIFT_ID": "DRIFT",
    "BREEZE_ID": "BREEZE",
    "WAVE_ID": "WAVE",
    "BLAZE_ID": "BLAZE",
    "BOLT_ID": "BOLT",
    "STARLIGHT_ID": "STARLIGHT",
    "MIST_ID": "MIST",
    "WHIRLWIND_ID": "WHIRLWIND",
    "DAWN_ID": "DAWN",
    "CRYSTAL_ID": "CRYSTAL",
    "LULLABY_ID": "LULLABY",
    "AURORA_ID": "AURORA",
    "RADIANCE_ID": "RADIANCE",
    "FLASH_ID": "FLASH"
}

# language IDs
LANGUAGES = {
    "ARABIC_ID": "ar",
    "CHINESE_ID": "zh-cn",
    "DANISH_ID": "da",
    "DUTCH_ID": "nl",
    "ENGLISH_ID": "en",
    "FRENCH_ID": "fr",
    "GERMAN_ID": "de",
    "HINDI_ID": "hi",
    "ICELANDIC_ID": "is",
    "ITALIAN_ID": "it",
    "JAPANESE_ID": "ja",
    "KOREAN_ID": "ko",
    "NORWEGIAN_ID": "nb",
    "POLISH_ID": "pl",
    "PORTUGUESE_BR_ID": "pt-br",
    "PORTUGUESE_ID": "pt",
    "ROMANIAN_ID": "ro",
    "RUSSIAN_ID": "ru",
    "SPANISH_ID": "es",
    "SPANISH_419_ID": "es-419",
    "SWEDISH_ID": "sv",
    "TURKISH_ID": "tr",
    "WELSH_ID": "cy"
}

VOICE_INFO =  {
    VOICES['ALTO_ID']: {"name": {"id": "text2speech.female", "default": "Alto", "description": "Name for a female voice."},
                    "gender": "female", "playbackRate": 1, "pitch": 1},
    VOICES['KITTEN_ID']: {"name": {"id": "text2speech.kitten", "default": "kitten", "description": "A baby cat."},
                    "gender": "female", "playbackRate": 1.41, "pitch": 1.5},
    VOICES['GIANT_ID']: {"name": {"id": "text2speech.giant", "default": "giant", "description": "A giant."},
                    "gender": "male", "playbackRate": 0.79, "pitch": 0.5},
    VOICES['TENOR_ID']: {"name": {"id": "text2speech.tenor", "default": "tenor", "description": "A tenor."},
                    "gender": "female", "playbackRate": 1.41, "pitch": 0.5},
    VOICES['ALIEN_ID']: {"name": {"id": "text2speech.alien", "default": "alien", "description": "An alien."},
                    "gender": "female", "playbackRate": 0.79, "pitch": 1.5},
    VOICES['THUNDER_ID']: {"name": {"id": "text2speech.thunder", "default": "Thunder", "description": "Male voice with low pitch and playback rate."},
                    "gender": "male", "playbackRate": 0.7, "pitch": 0.7},
    VOICES['STONE_ID']: {"name": {"id": "text2speech.stone", "default": "Stone", "description": "Male voice with low pitch."},
                    "gender": "male", "playbackRate": 1, "pitch": 0.7},
    VOICES['RUMBLE_ID']: {"name": {"id": "text2speech.rumble", "default": "Rumble", "description": "Male voice with high playback rate and low pitch."},
                    "gender": "male", "playbackRate": 1.3, "pitch": 0.7},
    VOICES['ECHO_ID']: {"name": {"id": "text2speech.echo", "default": "Echo", "description": "Male voice with standard pitch and low playback rate."},
                "gender": "male", "playbackRate": 0.7, "pitch": 1},
    VOICES['DRIFT_ID']: {"name": {"id": "text2speech.drift", "default": "Drift", "description": "Standard male voice."},
                    "gender": "male", "playbackRate": 1, "pitch": 1},
    VOICES['BREEZE_ID']: {"name": {"id": "text2speech.breeze", "default": "Breeze", "description": "Male voice with standard pitch and high playback rate."},
                    "gender": "male", "playbackRate": 1.3, "pitch": 1},
    VOICES['WAVE_ID']: {"name": {"id": "text2speech.wave", "default": "Wave", "description": "Male voice with high pitch and low playback rate."},
                "gender": "male", "playbackRate": 0.7, "pitch": 1.3},
    VOICES['BLAZE_ID']: {"name": {"id": "text2speech.blaze", "default": "Blaze", "description": "High-pitched male voice."},
                    "gender": "male", "playbackRate": 1, "pitch": 1.3},
    VOICES['BOLT_ID']: {"name": {"id": "text2speech.bolt", "default": "Bolt", "description": "Male voice with high pitch and playback rate."},
                "gender": "male", "playbackRate": 1.3, "pitch": 1.3},
    VOICES['STARLIGHT_ID']: {"name": {"id": "text2speech.starlight", "default": "Starlight", "description": "Female voice with low pitch and playback rate."},
                        "gender": "female", "playbackRate": 0.7, "pitch": 0.7},
    VOICES['MIST_ID']: {"name": {"id": "text2speech.mist", "default": "Mist", "description": "Female voice with low pitch."},
                "gender": "female", "playbackRate": 1, "pitch": 0.7},
    VOICES['WHIRLWIND_ID']: {"name": {"id": "text2speech.whirlwind", "default": "Whirlwind", "description": "Female voice with high playback rate and low pitch."},
                        "gender": "female", "playbackRate": 1.3, "pitch": 0.7},
    VOICES['DAWN_ID']: {"name": {"id": "text2speech.dawn", "default": "Dawn", "description": "Female voice with standard pitch and low playback rate."},
                "gender": "female", "playbackRate": 0.7, "pitch": 1},
    VOICES['CRYSTAL_ID']: {"name": {"id": "text2speech.crystal", "default": "Crystal", "description": "Standard female voice."},
                    "gender": "female", "playbackRate": 1, "pitch": 1},
    VOICES['LULLABY_ID']: {"name": {"id": "text2speech.lullaby", "default": "Lullaby", "description": "Female voice with standard pitch and high playback rate."},
                    "gender": "female", "playbackRate": 1.3, "pitch": 1},
    VOICES['AURORA_ID']: {"name": {"id": "text2speech.aurora", "default": "Aurora", "description": "Female voice with high pitch and low playback rate."},
                    "gender": "female", "playbackRate": 0.7, "pitch": 1.3},
    VOICES['RADIANCE_ID']: {"name": {"id": "text2speech.radiance", "default": "Radiance", "description": "High-pitched female voice."},
                    "gender": "female", "playbackRate": 1, "pitch": 1.3},
    VOICES['FLASH_ID']: {"name": {"id": "text2speech.flash", "default": "Flash", "description": "Female voice with high pitch and playback rate."},
                    "gender": "female", "playbackRate": 1.3, "pitch": 1.3}
    }
LANGUAGE_INFO = {
      LANGUAGES['ARABIC_ID']: {
        "name": "Arabic",
        "locales": ["ar"],
        "speechSynthLocale": "arb",
        "singleGender": True,
      },
      LANGUAGES['CHINESE_ID']: {
        "name": "Chinese (Mandarin)",
        "locales": ["zh-cn", "zh-tw"],
        "speechSynthLocale": "cmn-CN",
        "singleGender": True,
      },
      LANGUAGES['DANISH_ID']: {
        "name": "Danish",
        "locales": ["da"],
        "speechSynthLocale": "da-DK",
        "singleGender": False
      },
      LANGUAGES['DUTCH_ID']: {
        "name": "Dutch",
        "locales": ["nl"],
        "speechSynthLocale": "nl-NL",
        "singleGender": False
      },
      LANGUAGES['ENGLISH_ID']: {
        "name": "English",
        "locales": ["en"],
        "speechSynthLocale": "en-US",
        "singleGender": False
      },
      LANGUAGES['FRENCH_ID']: {
        "name": "French",
        "locales": ["fr"],
        "speechSynthLocale": "fr-FR",
        "singleGender": False
      },
      LANGUAGES['GERMAN_ID']: {
        "name": "German",
        "locales": ["de"],
        "speechSynthLocale": "de-DE",
        "singleGender": False
      },
      LANGUAGES['HINDI_ID']: {
        "name": "Hindi",
        "locales": ["hi"],
        "speechSynthLocale": "hi-IN",
        "singleGender": True,
      },
      LANGUAGES['ICELANDIC_ID']: {
        "name": "Icelandic",
        "locales": ["is"],
        "speechSynthLocale": "is-IS",
        "singleGender": False
      },
      LANGUAGES['ITALIAN_ID']: {
        "name": "Italian",
        "locales": ["it"],
        "speechSynthLocale": "it-IT",
        "singleGender": False
      },
      LANGUAGES['JAPANESE_ID']: {
        "name": "Japanese",
        "locales": ["ja", "ja-hira"],
        "speechSynthLocale": "ja-JP",
        "singleGender": False
      },
      LANGUAGES['KOREAN_ID']: {
        "name": "Korean",
        "locales": ["ko"],
        "speechSynthLocale": "ko-KR",
        "singleGender": True,
      },
      LANGUAGES['NORWEGIAN_ID']: {
        "name": "Norwegian",
        "locales": ["nb", "nn"],
        "speechSynthLocale": "nb-NO",
        "singleGender": True,
      },
      LANGUAGES['POLISH_ID']: {
        "name": "Polish",
        "locales": ["pl"],
        "speechSynthLocale": "pl-PL",
        "singleGender": False
      },
      LANGUAGES['PORTUGUESE_BR_ID']: {
        "name": "Portuguese (Brazilian)",
        "locales": ["pt-br"],
        "speechSynthLocale": "pt-BR",
        "singleGender": False
      },
      LANGUAGES['PORTUGUESE_ID']: {
        "name": "Portuguese (European)",
        "locales": ["pt"],
        "speechSynthLocale": "pt-PT",
        "singleGender": False
      },
      LANGUAGES['ROMANIAN_ID']: {
        "name": "Romanian",
        "locales": ["ro"],
        "speechSynthLocale": "ro-RO",
        "singleGender": True,
      },
      LANGUAGES['RUSSIAN_ID']: {
        "name": "Russian",
        "locales": ["ru"],
        "speechSynthLocale": "ru-RU",
        "singleGender": False
      },
      LANGUAGES['SPANISH_ID']: {
        "name": "Spanish (European)",
        "locales": ["es"],
        "speechSynthLocale": "es-ES",
        "singleGender": False
      },
      LANGUAGES['SPANISH_419_ID']: {
        "name": "Spanish (Latin American)",
        "locales": ["es-419"],
        "speechSynthLocale": "es-US",
        "singleGender": False
      },
      LANGUAGES['SWEDISH_ID']: {
        "name": "Swedish",
        "locales": ["sv"],
        "speechSynthLocale": "sv-SE",
        "singleGender": True,
      },
      LANGUAGES['TURKISH_ID']: {
        "name": "Turkish",
        "locales": ["tr"],
        "speechSynthLocale": "tr-TR",
        "singleGender": True,
      },
      LANGUAGES['WELSH_ID']: {
        "name": "Welsh",
        "locales": ["cy"],
        "speechSynthLocale": "cy-GB",
        "singleGender": True,
      },
    }

class Text2Speech:
    def __init__(self):
        self.SERVER_HOST = "https://synthesis-service.scratch.mit.edu"
        self.voiceSpeed = 1

    def speak(self, words, voice, language="en"):
        voice_id = voice.upper()
        # if voice_id not in VOICES.values():
        #     raise MartyCommandException(f"Voice must be one of {set(VOICES.values())}, not {voice_id}")
        # if language not in LANGUAGES.values():
        #     raise MartyCommandException(f"Language must be one of {set(LANGUAGES.values())}, not {language}")

        locale = self._get_speech_synth_locale(language)
        gender = VOICE_INFO[voice_id]["gender"]
        playback_rate = VOICE_INFO[voice_id]["playbackRate"] * self.voiceSpeed
        pitch = VOICE_INFO[voice_id]["pitch"]
        
        # Special case for voices where the synthesis service only provides a
        # single gender voice. In that case, always request the female voice,
        # and set special playback rates for the tenor and giant voices.
        if LANGUAGE_INFO[language]['singleGender']: 
            gender = "female"

        if voice_id == VOICES['KITTEN_ID']:
            words = " ".join(["meow" for word in words.split(" ")])
            locale = LANGUAGE_INFO["en"]["speechSynthLocale"]
        
        path = f"{self.SERVER_HOST}/synth"
        path += f"?locale={locale}"
        path += f"&gender={gender}"
        path += f"&text={quote(words[:128])}"    
        # perform http request to get audio file
        response = requests.get(path)
        response.raise_for_status()

        # Return the MP3 bytes as-is. Previously this method appended the
        # first bytes of the MP3 to the end to simulate a short silence, but
        # that caused the start of the phrase to repeat briefly.
        return response.content

    def _get_speech_synth_locale(self, language):
        if language in LANGUAGE_INFO:
            return LANGUAGE_INFO[language]["speechSynthLocale"]
        else:
            return "en"



import logging
import os
import time
import re
from typing import Callable, Dict, List, Optional, Union, Tuple
from packaging import version

from .ClientGeneric import ClientGeneric
from .RICCommsSerial import RICCommsSerial
from .RICCommsWiFi import RICCommsWiFi
from .RICCommsTest import RICCommsTest
from .RICProtocols import DecodedMsg, RICProtocols
from .RICROSSerial import RICROSSerial
from .RICInterface import RICInterface
from .RICHWElems import RICHWElems
from .Exceptions import (MartyConnectException,
                         MartyCommandException)
from .Text2Speech import Text2Speech

logger = logging.getLogger(__name__)

class ClientMV2(ClientGeneric):
    '''
    Lower level connector to Marty V2
    '''
    def __init__(self,
                method: str,
                locator: str,
                serialBaud: int = None,
                port = 80,
                wsPath = "/ws",
                subscribeRateHz = 10.0,
                ricInterface: Optional[RICInterface] = None,
                *args, **kwargs):
        '''
        Initialise connection to remote Marty
        Args:
            client_type: 'wifi' (for WiFi), 'usb' (for usb serial), 'exp' (for expansion serial),
                'test' (output is available via get_test_output())
            locator: str, ipAddress, hostname, serial-port, name of test file etc
                    depending on method
            serialBaud: serial baud rate
            port: IP port for websockets
            wsPath: path to use for websocket connection
            subscribeRateHz: rate of fastest subscription to events
        Raises:
            MartyConnectException if the connection to the host failed
        '''
        # Call base constructor
        super().__init__(*args, **kwargs)

        # Initialise vars
        self.subscribeRateHz = subscribeRateHz
        self.lastRICSerialMsgTime = None
        self.lastSubscrReqMsgTime = None
        self.maxTimeBetweenPubs = 10
        self.minTimeBetweenSubReqs = 10
        self.max_blocking_wait_time = 120  # seconds
        self.ricHardware = RICHWElems()
        self.isClosing = False
        self.ricSystemInfo = {}
        self.ricHwElemsInfoByIDNo = {}
        self.ricHwElemsList = []
        self.loggingCallback = None
        self.publishedMsgCallback = None
        self.reportMsgCallback = None
        self._initComplete = False
        self.maxWaitForConnReadySecs = 10
        self._minSysVersForSubscribeAPI = "1.0.0"
        self._interfaceMethod = method
        self._numHwStatusRetries = 10
        self._sendFileProgressCB = None
        self._playMP3ProgressCB = None
        # Debug
        self.DEBUG_RECEIVE_PUBLISHED_MSG = False
        self.DEBUG_RECEIVE_RICREST_MSGS = False
        self.DEBUG_GET_RIC_VERSION = False
        self.DEBUG_GET_HW_ELEMS_INFO = False
        self.DEBUG_SUBSCRIBE_TO_PUB_MSGS = False
        self.DEBUG_CONNECTION_PROCESS = False
        self.color_boundaries = {
            "red": ((20, 40, 100), (50, 300, 255)), 
            "yellow": ((85, 5, 25), (160, 100, 150)), 
            "green": ((180, 45, 55), (220, 230, 255)), 
            "blue": ((200, 0, 25), (320,  50, 150)), 
            "purple": ((345, 60, 40), (361, 200, 150)), 
            "red": ((0, 60, 40), (10, 200, 150)), 
        }
        # default calibration - suitable for batch 1 colour sensors that won't have saved calibration
        self.cal = [1.0, 1.65, 1.15, 1.0]

        # Check if we are given a RICInterface
        if ricInterface is None:
            # Handle the method of connection
            if method == "usb" or method == "exp":
                ifType = "overascii" if method == "usb" else "plain"
                if serialBaud is None:
                    serialBaud = 115200 if method == "usb" else 921600
                rifConfig = {
                    "serialPort": locator,
                    "serialBaud": serialBaud,
                    "ifType": ifType,
                }
                self.ricIF = RICInterface(RICCommsSerial())
            elif method == "test":
                rifConfig = {
                    "testFileName": locator
                }
                self.ricIF = RICInterface(RICCommsTest())
            else:
                rifConfig = {
                    "ipAddrOrHostname": locator,
                    "ipPort": port,
                    "wsPath": wsPath,
                    "ifType": "plain"
                }
                self.ricIF = RICInterface(RICCommsWiFi(self._onReconnect))
        else:
            rifConfig = {
                "ipAddrOrHostname": locator,
                "ifType": "plain"
            }
            self.ricIF = ricInterface

        # Debug
        ricConnStartTime = time.time()
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug("Starting to connect to RIC")

        # Open comms
        openOk = self.ricIF.open(rifConfig)
        if not openOk:
            raise MartyConnectException("Failed to open connection")

        # Debug
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug(f"RIC interface connection took {time.time() - ricConnStartTime} seconds")

        # Callbacks
        self.ricIF.setDecodedMsgCB(self._rxDecodedMsg)
        self.ricIF.setTimerCB(self._msgTimerCallback)
        self.ricIF.setLogLineCB(self._logDebugMsg)

    def start(self):
        # Debug
        debugRICOverallStart = time.time()
        debugStartTime = time.time()
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug("Starting to communicate with RIC")

        # Get the version of RIC
        self._getRICVersion()

        # Debug
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug(f"Got RIC version in {time.time() - debugStartTime} seconds")
        debugStartTime = time.time()

        # Get HWElems
        self._updateHwElemsInfo()

        # Debug
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug(f"Got HWElems in {time.time() - debugStartTime} seconds")

        # Now completed init
        self._initComplete = True

        # Debug
        # self._updateHwElemsInfo()

        # Wait for connection to be ready
        waitStartTime = time.time()
        while not self.is_conn_ready():
            if time.time() - waitStartTime > self.maxWaitForConnReadySecs:
                raise MartyConnectException("Connection to Marty not ready")
            time.sleep(0.1)

        # Debug
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug(f"Marty wait for ready time {time.time() - debugStartTime} seconds")

        # A flag is set on the first subscribed message - wait a little longer
        # to ensure that one of each subscribed message type has been received
        time.sleep(0.5)

        # Debug
        if self.DEBUG_CONNECTION_PROCESS:
            logger.debug(f"Marty overall start time {time.time() - debugRICOverallStart} seconds")

    def close(self):
        if self.isClosing:
            return
        self.isClosing = True
        # Send unsubscribe request
        self._unsubscribeFromPubMessages()
        # Close the RIC interface
        self.ricIF.close()

    def wait_if_required(self, expected_wait_ms: int, blocking_override: Union[bool, None]):
        if not self.is_blocking(blocking_override):
            return
        if self.subscribeRateHz <= 0:
            return

        deadline = time.time() + expected_wait_ms/1000 + self.max_blocking_wait_time
        time.sleep(2.5 * 1/self.subscribeRateHz)  # Give Marty time to report it is moving

        # The is_moving flag may be cleared briefly between 2 queued trajectories
        # Robot status may not be available for a while after Marty is turned on / connected to
        while self.is_moving() or self.get_robot_status().get('workQCount', 1) > 0:
            time.sleep(0.2 * 1/self.subscribeRateHz)
            if time.time() > deadline:
                raise TimeoutError("Marty wouldn't stop moving. Are you also controlling it via another method?"
                                   f"{os.linesep}If you issued some very long-running non-blocking commands, "
                                   "try increasing `marty.client.max_blocking_wait_time` (the current value "
                                   f"is {self.max_blocking_wait_time} s)")

    def hello(self) -> bool:
        return self.ricIF.cmdRICRESTRslt("traj/getReady")

    def get_ready(self) -> bool:
        return self.ricIF.cmdRICRESTRslt("traj/getReady")

    def stand_straight(self, move_time: int = 2000) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/standStraight?moveTime={move_time}")

    def discover(self) -> List[str]:
        return []

    def stop(self, stop_type: str, stopCode: int) -> bool:
        robotRestCmd = "stop"
        trajCmd = ""
        if stopCode == 0: robotRestCmd = "stopAfterMove"
        elif stopCode == 2: robotRestCmd = "panic"
        elif stopCode == 3: trajCmd = "getReady"
        elif stopCode == 4 or stopCode == 5: robotRestCmd = "pause"
        isOk = self.ricIF.cmdRICRESTRslt("robot/" + robotRestCmd)
        if len(trajCmd) > 0:
            self.ricIF.cmdRICRESTRslt("traj/" + trajCmd)
        return isOk

    def resume(self) -> bool:
        return self.ricIF.cmdRICRESTRslt("robot/resume")

    def hold_position(self, hold_time: int) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/hold?moveTime={hold_time}")

    def move_joint(self, joint_id: int, position: int, move_time: int) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/joint?jointID={joint_id}&angle={position}&moveTime={move_time}")

    def get_joint_position(self, joint_id: Union[int, str]) -> float:
        return self.ricHardware.getServoPos(joint_id, self.ricHwElemsInfoByIDNo)

    def get_joint_current(self, joint_id: Union[int, str]) -> float:
        return self.ricHardware.getServoCurrent(joint_id, self.ricHwElemsInfoByIDNo)

    def get_joint_status(self, joint_id: Union[int, str]) -> int:
        return self.ricHardware.getServoFlags(joint_id, self.ricHwElemsInfoByIDNo)

    def lean(self, direction: str, amount: Optional[int], move_time: int) -> bool:
        if amount is None:
            amount = 29
        try:
            directionNum = ClientGeneric.SIDE_CODES[direction]
        except KeyError:
            self.preException(True)
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(ClientGeneric.SIDE_CODES.keys()), direction))
        return self.ricIF.cmdRICRESTRslt(f"traj/lean?side={directionNum}&leanAngle={amount}&moveTime={move_time}")

    def walk(self, num_steps: int = 2, start_foot:str = 'auto', turn: int = 0,
                step_length:int = 25, move_time: int = 1500) -> bool:
        side_url_param = ''
        if start_foot != 'auto':
            try:
                sideNum = ClientGeneric.SIDE_CODES[start_foot]
            except KeyError:
                raise MartyCommandException("Direction must be one of {}, not '{}'"
                                            "".format(set(ClientGeneric.SIDE_CODES.keys()), start_foot))
            side_url_param = f'&side={sideNum}'
        return self.ricIF.cmdRICRESTRslt(f"traj/step/{num_steps}?stepLength={step_length}&turn={turn}"
                                         f"&moveTime={move_time}" + side_url_param)

    def lift_foot(self, side: str) -> bool:
        try:
            sideNum = ClientGeneric.SIDE_CODES[side]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(ClientGeneric.SIDE_CODES.keys()), side))
        return self.ricIF.cmdRICRESTRslt(f"traj/liftFoot/1/?side={sideNum}")

    def lower_foot(self, side: str) -> bool:
        try:
            sideNum = ClientGeneric.SIDE_CODES[side]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(ClientGeneric.SIDE_CODES.keys()), side))
        return self.ricIF.cmdRICRESTRslt(f"traj/lowerFoot/1/?side={sideNum}")

    def wave(self, side: str = 'right') -> bool:
        try:
            sideNum = ClientGeneric.SIDE_CODES[side]
        except KeyError:
            raise MartyCommandException("Direction must be one of {}, not '{}'"
                                        "".format(set(ClientGeneric.SIDE_CODES.keys()), side))
        return self.ricIF.cmdRICRESTRslt(f"traj/wave/1/?side={sideNum}")

    def eyes(self, joint_id: int, pose_or_angle: Union[str, int], move_time: int = 100) -> bool:
        if type(pose_or_angle) is str:
            try:
                eyesTrajectory = ClientGeneric.EYE_POSES[pose_or_angle]
            except KeyError:
                raise MartyCommandException("pose must be one of {}, not '{}'"
                                            "".format(set(ClientGeneric.EYE_POSES.keys()), pose_or_angle))
            return self.ricIF.cmdRICRESTRslt(f"traj/{eyesTrajectory}")
        return self.move_joint(joint_id, pose_or_angle, move_time)

    def kick(self, side: str = 'right', twist: int = 0, move_time: int = 2500) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/kick?side={ClientGeneric.SIDE_CODES[side]}&moveTime={move_time}&turn={twist}")

    def arms(self, left_angle: int, right_angle: int, move_time: int) -> bool:
        self.move_joint(6, left_angle, move_time)
        return self.move_joint(7, right_angle, move_time)

    def celebrate(self, move_time: int = 4000) -> bool:

        # TODO - add celebrate trajectory to Marty V2
        return self.ricIF.cmdRICRESTRslt(f"traj/wiggle?moveTime={move_time}")

    def circle_dance(self, side: str = 'right', move_time: int = 2500) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/circle?side={ClientGeneric.SIDE_CODES[side]}&moveTime={move_time}")

    def dance(self, side: str = 'right', move_time: int = 3000) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/dance?side={ClientGeneric.SIDE_CODES[side]}&moveTime={move_time}")

    def wiggle(self, move_time: int = 4000) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"traj/wiggle?moveTime={move_time}")

    def sidestep(self, side: str, steps: int = 1, step_length: int = 35,
            move_time: int = 1000) -> bool:
        if side != 'right' and side != 'left':
            raise MartyCommandException("side must be one of 'right' or 'left', not '{}'"
                                        "".format(side))
        return self.ricIF.cmdRICRESTRslt(f"traj/sidestep/{steps}?side={ClientGeneric.SIDE_CODES[side]}"
                                         f"&stepLength={step_length}&moveTime={move_time}")

    def set_volume(self, volume: int) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"audio/vol/{volume}")

    def get_volume(self) -> int:
        result = self.ricIF.cmdRICRESTURLSync("audio/vol")
        if result.get("rslt", "") == "ok":
            return result.get("volPC", 0)
        return 0

    def play_sound(self, name_or_freq_start: Union[str,float],
            freq_end: Optional[float] = None,
            duration: Optional[int] = None) -> bool:
        if name_or_freq_start.lower().endswith(".mp3"):
            return self.ricIF.play_mp3(name_or_freq_start)
        if not name_or_freq_start.lower().endswith(".raw"):
            name_or_freq_start += ".raw"
        return self.ricIF.cmdRICRESTRslt(f"filerun/{name_or_freq_start}")

    def pinmode_gpio(self, gpio: int, mode: str) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def write_gpio(self, gpio: int, value: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def digitalread_gpio(self, gpio: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def i2c_write(self, *byte_array: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def i2c_write_to_ric(self, address: int, byte_array: bytes) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_battery_voltage(self) -> float:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_battery_remaining(self) -> float:
        powerStatus = self.ricHardware.getPowerStatus()
        return powerStatus.get("battRemainCapacityPercent", 0)

    def get_distance_sensor(self) -> int:
        for attached_add_on in self.get_add_ons_status().values():
            if type(attached_add_on) == dict and attached_add_on['whoAmI'] == 'VCNL4200':
                distance_bytes = attached_add_on['data'][1:3]
                distance = int.from_bytes(distance_bytes, 'big')
                return distance
        return 0

    def _data_color(self, attached_add_on: dict) -> Tuple[int, str]:
        '''
        Parses data of passed-in color sensor :two:
        Args:
            attached_add_on: Name of color sensor add on
        Returns:
            A tuple with four integers of data (clear, red, green, blue)
            and a string of whether the sensor is a default 'left' or 'right' add on
        '''
        if attached_add_on['whoAmI'] == 'coloursensor':
            clear = attached_add_on['data'][1] * self.cal[0]
            red = attached_add_on['data'][2] * self.cal[1]
            green = attached_add_on['data'][3] * self.cal[2]
            blue = attached_add_on['data'][4] * self.cal[3]
            return (clear, red, green, blue), 'left'
        raise MartyCommandException(
            f"The add on '{attached_add_on['name']}' is not a valid add on for color sensing."
            "Please make sure to pass in the name of a color sensor."
        )

    def _index_data_color_ir(self, attached_add_on: str) -> Tuple[Tuple[int, int, int], str]:
        '''
        Parses data of passed-in color or IR sensor :two:
        Args:
            attached_add_on: Name of IR or color sensor add on
        Returns:
            A tuple with three integers of data (detection byte, obstacle value, ground value)
            and a string of whether the sensor is a default 'left' or 'right' add on
        '''
        if attached_add_on['whoAmI'] == 'IRFoot':
            detection_flags = attached_add_on['data'][1]
            obstacle_data_raw = int.from_bytes(attached_add_on['data'][2:4], 'big')
            ground_data_raw = int.from_bytes(attached_add_on['data'][4:6], 'big')
            return (detection_flags, obstacle_data_raw, ground_data_raw), 'right'
        if attached_add_on['whoAmI'] == 'coloursensor':
            detection_flags = attached_add_on['data'][6]
            obstacle_data_raw = int.from_bytes(attached_add_on['data'][8:10], 'big')
            ground_data_raw = attached_add_on['data'][2]
            return (detection_flags, obstacle_data_raw, ground_data_raw), 'left'

    def _get_obstacle_and_ground_raw_data(self, add_on_or_side: str) -> list:
        '''
        Finds the wanted color or IR sensor and gets the parsed obstacle and ground data :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor, name of an IR
            sensor, `'left'` for the add on connected to the left foot,
            or `'right'` for the add on connected to the right foot
        Returns:
            A tuple of the detection byte, obstacle reading, and ground reading of the add on
        '''
        sensor_whoamis = {'IRFoot', 'coloursensor'}
        sensor_data = {'left': [], 'right': []}
        sensor_possible_names = {
            'left': ['LeftColorSensor', 'LeftIRFoot'],
            'right': ['RightIRFoot', 'RightColorSensor']
        }
        addon_names = sensor_possible_names.get(add_on_or_side, [])

        if addon_names == []:
            addon_names = [add_on_or_side]

        # There may be a situation just after connecting to RIC where publication messages from the addon
        # have not yet been received so we need to wait for them to be received before we can parse them
        for retryLoop in range(10):
            addon_statuses = self.get_add_ons_status().values()
            if len(addon_statuses) != 0:
                break
            time.sleep(0.1)

        # Get the add-on values
        for attached_add_on in addon_statuses:
            if type(attached_add_on) == dict and attached_add_on.get('whoAmI','') in sensor_whoamis:
                obstacle_and_ground_data, side = self._index_data_color_ir(attached_add_on)
                if attached_add_on['name'] in addon_names:
                    return obstacle_and_ground_data
                else:
                    sensor_data[side].append(obstacle_and_ground_data)
        if len(sensor_data[add_on_or_side.lower()]) == 1:
            return sensor_data[add_on_or_side.lower()][0]
        elif add_on_or_side.lower() == 'left' or add_on_or_side.lower() == 'right':
            raise MartyCommandException(
                f"Marty could not find a {add_on_or_side} sensor. "
                "Please make sure your add ons are plugged in and named correctly"
            )
        else:
            raise MartyCommandException(
                f"The add on '{add_on_or_side}' is not a valid add on for obstacle and ground sensing."
                "Please make sure to pass in the name of an IR sensor or Color sensor."
            )

    def _get_color_sensor_raw_data(self, add_on_or_side: str) -> list:
        '''
        Finds the wanted color sensor and gets the parsed color data :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor,
            `'left'` for the add on connected to the left foot,
            or `'right'` for the add on connected to the right foot
        Returns:
            A tuple of the detection byte, obstacle reading, and ground reading of the add on
        '''
        sensor_whoamis = {'coloursensor'}
        sensor_data = {'left': [], 'right': []}
        sensor_possible_names = {
            'left': ['LeftColorSensor'],
            'right': ['RightColorSensor']
        }
        addon_names = sensor_possible_names.get(add_on_or_side, [])

        if addon_names == []:
            addon_names = [add_on_or_side]

        # There may be a situation just after connecting to RIC where publication messages from the addon
        # have not yet been received so we need to wait for them to be received before we can parse them
        for retryLoop in range(10):
            addon_statuses = self.get_add_ons_status().values()
            if len(addon_statuses) != 0:
                break
            time.sleep(0.1)

        # Get the add-on values
        for attached_add_on in addon_statuses:
            if type(attached_add_on) == dict and attached_add_on.get('whoAmI','') in sensor_whoamis:
                obstacle_and_ground_data, side = self._index_data_color_ir(attached_add_on)
                color_data, side = self._data_color(attached_add_on)
                if attached_add_on['name'] in addon_names:
                    return (obstacle_and_ground_data, color_data)
                else:
                    sensor_data[side].append(obstacle_and_ground_data)
        if len(sensor_data[add_on_or_side.lower()]) == 1:
            return sensor_data[add_on_or_side.lower()][0]
        elif add_on_or_side.lower() == 'left' or add_on_or_side.lower() == 'right':
            raise MartyCommandException(
                f"Marty could not find a {add_on_or_side} sensor. "
                "Please make sure your add ons are plugged in and named correctly"
            )
        else:
            raise MartyCommandException(
                f"The add on '{add_on_or_side}' is not a valid add on for obstacle and ground sensing."
                "Please make sure to pass in the name of an IR sensor or Color sensor."
            )


    def foot_on_ground(self, add_on_or_side: str) -> bool:
        raw_data = self._get_obstacle_and_ground_raw_data(add_on_or_side)
        if len(raw_data) > 0:
            return (raw_data[0] & 0b10) == 0b00
        return False

    def foot_obstacle_sensed(self, add_on_or_side: str) -> bool:
        raw_data = self._get_obstacle_and_ground_raw_data(add_on_or_side)
        if len(raw_data) > 0:
            return (raw_data[0] & 0b1) == 0b1
        return False

    def get_color_sensor_color(self, add_on_or_side: str) -> int:
        '''
        Gets the colour detected by a colour sensor :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            one of  "yellow", "green", "blue", "purple", "red", "air", "unclear"
        '''
        [foot_on_ground, raw_data] = self._get_color_sensor_raw_data(add_on_or_side) # (clear, red, green, blue)
        if len(raw_data) > 0:
            foot_on_ground = (foot_on_ground[0] & 0b10) == 0b00
            if not foot_on_ground:
                return "air"
            [hue, chroma] = self._get_hue_color(raw_data[1], raw_data[2], raw_data[3])
            for color, (lower, upper) in self.color_boundaries.items():
                if lower[0] <= hue <= upper[0] and lower[1] <= chroma <= upper[1]:
                    return color
            return "unclear"
        return 0

    def get_color_sensor_hex(self, add_on_or_side: str) -> str:
        '''
        Gets the colour detected by a colour sensor :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
        Returns:
            The hex code of the colour detected by the sensor
        '''
        [foot_on_ground, raw_data] = self._get_color_sensor_raw_data(add_on_or_side) # (clear, red, green, blue)
        if len(raw_data) > 0:
            hex_color = self.rgb_to_hex((raw_data[1], raw_data[2], raw_data[3]))
            return hex_color
        return 0
    
    def get_color_sensor_value_by_channel(self, add_on_or_side: str, channel: str) -> int:
        '''
        Gets the value of a colour sensor channel :two:
        Args:
            add_on_or_side: Takes in the name of a color sensor `'left'` for the add on connected to the left foot,
             or `'right'` for the add on connected to the right foot. 
            channel: Takes in the name of a channel `'clear'` `'red'`, `'green'`, or `'blue'`
        Returns:
            The value of the channel
        '''
        [foot_on_ground, raw_data] = self._get_color_sensor_raw_data(add_on_or_side)
        if len(raw_data) > 0:
            if channel.lower() == 'clear':
                return raw_data[0]
            elif channel.lower() == 'red':
                return raw_data[1]
            elif channel.lower() == 'green':
                return raw_data[2]
            elif channel.lower() == 'blue':
                return raw_data[3]
            else:
                raise MartyCommandException(
                    f"The channel '{channel}' is not a valid channel. "
                    "Please make sure to pass in the name of a valid channel."
                )
        return 0
    
    def speak(self, words: str = "hello", voice: str = "alto") -> bool:
        text2speech = Text2Speech()
        audio = text2speech.speak(words, voice)
        # store audio data to a file and then load it up and play it
        with open("temp.mp3", "wb") as f:
            f.write(audio)  

        result = self.ricIF.streamSoundFile("temp.mp3", "streamaudio", self._playMP3ProgressAdapter)
        os.remove("temp.mp3")
        return result


    def get_obstacle_sensor_reading(self, add_on_or_side: str) -> int:
        raw_data = self._get_obstacle_and_ground_raw_data(add_on_or_side)
        if len(raw_data) > 1:
            return raw_data[1]
        return 0

    def get_ground_sensor_reading(self, add_on_or_side: str) -> int:
        raw_data = self._get_obstacle_and_ground_raw_data(add_on_or_side)
        if len(raw_data) > 2:
            return raw_data[2]
        return 0

    def get_accelerometer(self, axis: Optional[str] = None, axisCode: int = 0) -> float:
        if axis is None:
            return self.ricHardware.getIMUAll()
        return self.ricHardware.getIMUAxisValue(axisCode)

    def enable_motors(self, enable: bool = True, clear_queue: bool = True) -> bool:
        return True

    def enable_safeties(self, enable: bool = True) -> bool:
        return True

    def fall_protection(self, enable: bool = True) -> bool:
        return True

    def motor_protection(self, enable: bool = True) -> bool:
        return True

    def battery_protection(self, enable: bool = True) -> bool:
        return True

    def buzz_prevention(self, enable: bool = True) -> bool:
        return True

    def lifelike_behaviour(self, enable: bool = True) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def set_parameter(self, *byte_array: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def save_calibration(self) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"calibrate/set")

    def clear_calibration(self) -> bool:
        return self.ricIF.cmdRICRESTRslt(f"calibrate/setFlag/0")

    def is_calibrated(self) -> bool:
        result = self.ricIF.cmdRICRESTURLSync("calibrate")
        if result.get("rslt", "") == "ok":
            return result.get("calDone", 0) != 0
        return False

    def ros_command(self, *byte_array: int) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def keyframe (self, time: float, num_of_msgs: int, msgs) -> List[bytes]:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_chatter(self) -> bytes:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def get_firmware_version(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def _mute_serial(self) -> bool:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def ros_serial_formatter(self, topicID: int, send: bool = False, *message: int) -> List[int]:
        raise MartyCommandException(ClientGeneric.NOT_IMPLEMENTED)

    def is_moving(self) -> bool:
        return self.ricHardware.getIsMoving()

    def is_paused(self) -> bool:
        return self.ricHardware.getIsPaused()

    def get_robot_status(self) -> Dict:
        return self.ricHardware.getRobotStatus()

    def get_joints(self) -> Dict:
        return self.ricHardware.getServos(self.ricHwElemsInfoByIDNo)

    def get_power_status(self) -> Dict:
        return self.ricHardware.getPowerStatus()

    def get_add_ons_status(self) -> Dict:
        return self.ricHardware.getAddOns(self.ricHwElemsInfoByIDNo)

    def get_add_on_status(self, add_on_name_or_id: Union[int, str]) -> Dict:
        return self.ricHardware.getAddOn(add_on_name_or_id, self.ricHwElemsInfoByIDNo)

    def add_on_query(self, add_on_name: str, data_to_write: bytes, num_bytes_to_read: int) -> Dict:
        return self.ricIF.addOnQueryRaw(add_on_name, data_to_write, num_bytes_to_read)

    def get_system_info(self) -> Dict:
        return self.ricSystemInfo

    def set_marty_name(self, name: str) -> bool:
        escapedName = name.replace('"', '').replace('\n','')
        return self.ricIF.cmdRICRESTRslt(f"friendlyname/{escapedName}")

    def get_marty_name(self) -> str:
        result = self.ricIF.cmdRICRESTURLSync("friendlyname")
        if result.get("rslt", "") == "ok":
            return result.get("friendlyName", "Marty")
        return "Marty"

    def is_marty_name_set(self) -> bool:
        result = self.ricIF.cmdRICRESTURLSync("friendlyname")
        if result.get("rslt", "") == "ok":
            return result.get("friendlyNameIsSet", 0) != 0
        return False

    def get_hw_elems_list(self) -> List:
        self._updateHwElemsInfo()
        return self.ricHwElemsList

    def send_ric_rest_cmd(self, ricRestCmd: str) -> None:
        self.ricIF.sendRICRESTURL(ricRestCmd)

    def send_ric_rest_cmd_sync(self, ricRestCmd: str) -> Dict:
        return self.ricIF.cmdRICRESTURLSync(ricRestCmd)

    def is_conn_ready(self) -> bool:
        if not self.ricIF.isOpen():
            return False
        return self._initComplete and ((self.lastRICSerialMsgTime is not None) or (self.subscribeRateHz == 0) or (len(self.ricHwElemsList) < 10))

    def _is_valid_disco_addon(self, add_on: str) -> bool:
        disco_whoamis = {"LEDfoot", "LEDarm", "LEDeye"}
        for attached_add_on in self.get_add_ons_status().values():
            print("attached_add_on", attached_add_on)
            if type(attached_add_on) == dict and attached_add_on['name'] == add_on:
                if attached_add_on['whoAmI'] in disco_whoamis:
                    return True
                else:
                    raise MartyCommandException(f"The add on name: '{add_on}' is not a valid disco add on. "
                                                "Please check the add on name in the scratch app -> configure -> add ons")
        raise MartyCommandException(f"The add on name '{add_on}' is not a valid add on. Please check the add on "
                                    "name in the scratch app -> configure -> add ons")

    def disco_off(self, add_on: str, api = 'led') -> bool:
        if api == 'raw_query':
            if self._is_valid_disco_addon(add_on):
                response = self.add_on_query(add_on, bytes.fromhex('01'), 0)
                return response.get("rslt", "") == "ok"
        elif api == 'led':
            return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/off")

    def disco_pattern(self, pattern: int, add_on: str) -> bool:
        if pattern == 1:
            pattern = '10'
        elif pattern == 2:
            pattern = '11'
        else:
            raise Exception("Pattern must be 1 or 2")
        if self._is_valid_disco_addon(add_on):
            response = self.add_on_query(add_on, bytes.fromhex(pattern), 0)
            return response.get("rslt", "") == "ok"
    
    def disco_named_pattern(self, add_on: str, pattern: str) -> bool:
        try :
            pattern = ClientGeneric.DISCO_PATTERNS[pattern]
        except KeyError:
            raise MartyCommandException(f"Pattern must be one of {set(ClientGeneric.DISCO_PATTERNS.keys())}, not '{pattern}'")
        if (pattern == ClientGeneric.DISCO_PATTERNS["off"]):
            return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/{pattern}")
        return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/pattern/{pattern}")

    def _region_to_bytes(self, region: Union[str, int]) -> bytes:
        if region == 'all':
            region = (2,)
        else:
            region = (4, region)
        return bytes(region)

    def _downscale_color(self, color: Union[tuple, bytes]):
        return bytes(c//25 for c in color)

    def _is_valid_color_hex(self, color_hex: str) -> bool:
        hex_pattern = re.compile("^([A-Fa-f0-9]{6})$")
        return bool(hex_pattern.match(color_hex))

    def _parse_color_hex(self, color_hex: str, region: Union[str, int]) -> bytes:
        input_color = color_hex
        color_hex = color_hex.lstrip('#')
        if self._is_valid_color_hex(color_hex):
            color_bytes = bytes.fromhex(color_hex)
        else:
            raise MartyCommandException(f"The string '{input_color}' is not a valid hex color code or default color")
        return color_bytes

    def disco_color(self, color: Union[str, Tuple[int, int, int]], add_on: str, region: Union[int, str]) -> bool:
        default_colors = {
            'white'  : 'FFFFFF',
            'red'    : 'FF0000',
            'blue'   : '0000FF',
            'yellow' : 'FFFF00',
            'green'  : '008000',
            'teal'   : '008080',
            'pink'   : 'eb1362',
            'purple' : '7800c8',
            'orange' : '961900'
        }
        if type(color) is str:
            color = default_colors.get(color.lower(), color)
            color = self._parse_color_hex(color, region)
        elif type(color) is tuple:
            if len(color) != 3:
                raise MartyCommandException(f'RGB tuple must be 3 numbers, instead of: {color}. Please enter a valid color.')
        else:
            raise MartyCommandException(f"Color must be of string or tuple form, not {type(color)}")
        color = self._downscale_color(color)
        region = self._region_to_bytes(region)
        command = region + color
        if self._is_valid_disco_addon(add_on):
            response = self.add_on_query(add_on, command, 0)
            return response.get("rslt", "") == "ok"

    def disco_group_operation(self, disco_operation: Callable, whoamis: set, operation_kwargs: dict, api='led') -> bool:
        '''
        Calls disco operations in groups for multiple add ons :two:
        Args:
            disco_operation: function for disco add on
            whoami_type_codes: the add ons that the function applies to
            operation_kwargs: additional arguments that need to be passed into the operation
            api: the api that the function uses (either 'raw_query' or 'led')
        Returns:
            True if Marty accepted all requests
        '''
        result = True
        # if any of the whoamis is equal to "00000089" then 
        # we need to also send to LEDeye in case the new LEDeyes (batch 4) are connected
        if "00000089" in whoamis and api == 'led':
            result = result and disco_operation(add_on="LEDeye", **operation_kwargs)
        for attached_add_on in self.get_add_ons_status().values():
            if type(attached_add_on) == dict and attached_add_on['whoAmI'] in whoamis:
                addon_name = attached_add_on['name']
                result = result and disco_operation(add_on=addon_name, **operation_kwargs)
        return result

    def disco_color_led_api(self, color: Union[str, Tuple[int, int, int]], add_on: str, region: Union[int, str]) -> bool:
        '''
        Sets the color of a disco add on :two:
        Args:
            color: RGB tuple
            add_on: name of disco add on
            region: region of add on to set color
        Returns:
            True if Marty accepted request
        '''
        default_colors = {
            'white'  : 'FFFFFF',
            'red'    : 'FF0000',
            'blue'   : '0000FF',
            'yellow' : 'FFFF00',
            'green'  : '008000',
            'teal'   : '008080',
            'pink'   : 'eb1362',
            'purple' : '7800c8',
            'orange' : '961900'
        }
        if type(color) is str:
            color = default_colors.get(color.lower(), color)
        elif type(color) is tuple:
            if len(color) != 3:
                raise MartyCommandException(f'RGB tuple must be 3 numbers, instead of: {color}. Please enter a valid color.')
            color = self.rgb_to_hex(color)
        if region == 'all':
            return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/color/{color}")
        else:
            print(f"led/{add_on}/color/region/{region}/{color}")
            return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/region/{region}/{color}")

    def disco_color_specific_led(self, color: Union[str, Tuple[int, int, int]], add_on: str, add_on_who_am_i: str, led_id: int) -> bool:
        '''
        Sets the color of specific LED light of a disco add on :two:
        Args:
            color: RGB tuple
            add_on: name of disco add on
            add_on_who_am_i: who am i string of the add on
            led_id: ID of LED to set color
        Returns:
            True if Marty accepted request
        '''
        default_colors = {
            'white'  : 'FFFFFF',
            'red'    : 'FF0000',
            'blue'   : '0000FF',
            'yellow' : 'FFFF00',
            'green'  : '008000',
            'teal'   : '008080',
            'pink'   : 'eb1362',
            'purple' : '7800c8',
            'orange' : '961900'
        }
        if type(color) is str:
            color = default_colors.get(color.lower(), color)
        elif type(color) is tuple:
            if len(color) != 3:
                raise MartyCommandException(f'RGB tuple must be 3 numbers, instead of: {color}. Please enter a valid color.')
            color = self.rgb_to_hex(color)
        led_id = self.led_id_mapping(id=led_id, is_from_color_picker=False, add_on_who_am_i=add_on_who_am_i)
        return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/setled/{led_id}/{color}")

    def disco_color_eyepicker(self, colours: Union[str, List[str]], add_on: str) -> bool:
        '''
        Turn on disco add on specific LED lights to specific colors :two:
        Args:
            colours: list of colors to switch the LEDs to; takes in a hex code, the position in the list corresponds to the LED ID
            add_on: add on name of which the function applies to
        Returns:
            True if Marty accepted the request
        '''
        if type(colours) is str:
            colours = [colours]
            colours = ['{}'.format(colour.replace("#", "")) for colour in colours]
        elif type(colours) is list:
            colours = ['{}'.format(colour.replace("#", "")) for colour in colours]
            for colour in colours:
                if type(colour) is not str:
                    raise MartyCommandException(f'Colour must be a string, not {type(colour)}')
                if len(colour) != 6:
                    raise MartyCommandException(f'Colour must be a hex code, not {colour}')
        if len(colours) > 12:
            raise MartyCommandException(f'Colours must be equal to or less than 12, not {len(colours)}')
        return self.ricIF.cmdRICRESTRslt(f"led/{add_on}/color/persist?{'&'.join([f'{self.led_id_mapping(id=i, is_from_color_picker=True)}={colour}' for i, colour in enumerate(colours)])}")

    def function_led(self, colour: Tuple[int, int, int], breathe: str = "on", breath_ms: int = 1000) -> bool:
        '''
        Turn on function LED light to a color :two:
        Args:
            colour: color to switch the LEDs to; takes in an RGB tuple of integers between 0-255
            breathe: "on"/"breath"
            breath_ms: time in milliseconds for the LED to breath
        Returns:
            True if Marty accepted the request
        '''
        if len(colour) != 3:
            raise MartyCommandException(f'RGB tuple must be 3 numbers, instead of: {colour}. Please enter a valid color.')
        print(f"indicator/set?pixIdx=1;blinkType=on;r={colour[0]};g={colour[1]};b={colour[2]};rateMs={breath_ms}")
        return self.ricIF.cmdRICRESTRslt(f"indicator/set?pixIdx=1;blinkType={breathe};r={colour[0]};g={colour[1]};b={colour[2]};rateMs={breath_ms}")

    def function_led_off(self) -> bool:
        '''
        Turn off function LED light :two:
        Returns:
            True if Marty accepted the request
        '''
        return self.ricIF.cmdRICRESTRslt(f"indicator/set?pixIdx=1;blinkType=off;r=0;g=0;b=0;rateMs=1000")

    def register_logging_callback(self, loggingCallback: Callable[[str],None]) -> None:
        '''
        Register a callback function to be called on every log message from RIC.
        Log messages are used mainly for debugging and error reporting.
        Args:
            messageCallback: a callback function (with one string argument - the log message)
        Returns:
            None
        '''
        self.loggingCallback = loggingCallback

    def register_publish_callback(self, messageCallback: Callable[[int],None]) -> None:
        '''
        Register a callback function to be called on every message published by RIC.
        RIC publishes information like the accelerometer and joint positions constantly.
        If registered, the callback is called after the message is fully decoded, so a common
        use-case is to check the topic (the int passed to the callback) to see if the information
        is of interest (for example topic == Marty.PUBLISH_TOPIC_ACCELEROMETER if new accelerometer
        data is available). Then get the changed information using the regular get_accelerometer method
        (or get_joints, etc for other data).
        Args:
            messageCallback: a callback function (with one argument - the topic code of the published
                    message) that will be called on every message published by RIC.
        Returns:
            None
        '''
        self.publishedMsgCallback = messageCallback

    def register_report_callback(self, messageCallback: Callable[[str],None]) -> None:
        '''
        Register a callback function to be called on every report message recieved from RIC.
        Report messages are used for alert conditions such as Falling, Over Current, etc.
        The report message (passed in the callback function str) is a JSON string.
        The elements of the JSON string include:
            msgType: the type of report message (this will be "warn" for all alerts)
            msgBody: the message text (this is "freeFallDet" or "overCurrentDet")
            IDNo: the IDNo of the element that generated the report (the codes for these
                are in the Marty.HW_ELEM_IDS dictionary but additional ones may appear if
                add-ons create warnings - the IDNos of add-ons are not fixed numbers but you
                can see the IDNo of add-ons by using the get_add_ons_status() method)
        Args:
            messageCallback: a callback function (with one string argument)
                    that will be called on every message published by RIC.
        Returns:
            None
        '''
        self.reportMsgCallback = messageCallback

    def get_interface_stats(self) -> Dict:
        ricIFStats = self.ricIF.getStats()
        publishInfo = self.ricHardware.getPublishStats()
        ricIFStats.update(publishInfo)
        return ricIFStats

    def preException(self, isFatal: bool) -> None:
        if isFatal:
            self.ricIF.close()
        logger.warning(f"Pre-exception isFatal {isFatal}")

    def _rxDecodedMsg(self, decodedMsg: DecodedMsg, interface: RICInterface):
        if decodedMsg.protocolID == RICProtocols.PROTOCOL_ROSSERIAL:
            if self.DEBUG_RECEIVE_PUBLISHED_MSG:
                logger.debug(f"ROSSERIAL message received {len(decodedMsg.payload)}")
            self.lastRICSerialMsgTime = time.time()
            if decodedMsg.payload:
                RICROSSerial.decode(decodedMsg.payload, 0, self._rxPublishedMsg)
        elif decodedMsg.protocolID == RICProtocols.PROTOCOL_RICREST:
            if self.DEBUG_RECEIVE_RICREST_MSGS:
                logger.debug(f"RICREST message received {decodedMsg.payload}")
            # Callback for REPORT messages
            if (self.reportMsgCallback is not None) and (decodedMsg.msgTypeCode == RICProtocols.MSG_TYPE_REPORT):
                try:
                    self.reportMsgCallback(decodedMsg.payload)
                except:
                    logger.exception("Report callback failed:")
        else:
            logger.info(f"RIC UNKNOWN message received {decodedMsg.payload}")
            pass

    def _rxPublishedMsg(self, topicID: int, payload: bytes):
        # Decode the message
        self.ricHardware.updateWithROSSerialMsg(topicID, payload)
        # Callback on published messages
        if self.publishedMsgCallback is not None:
            try:
                self.publishedMsgCallback(topicID)
            except:
                logger.exception("Publish callback failed:")

    def _logDebugMsg(self, logMsg: str) -> None:
        if self.loggingCallback:
            try:
                self.loggingCallback(logMsg)
            except:
                logger.exception("Logging callback failed:")

    def _msgTimerCallback(self) -> None:
        if self.isClosing:
            return
        self._subscribeToPubMessages(False)

    def _getRICVersion(self) -> bool:
        # Retries here to allow for baud-rate changes, etc
        for retries in range(self._numHwStatusRetries):
            # logger.debug(f"_getRICVersion attempt {retries+1}")
            self.ricSystemInfo = self.ricIF.cmdRICRESTURLSync("v")
            if self.ricSystemInfo.get("rslt", "") == "ok":
                break
        if self.DEBUG_GET_RIC_VERSION:
            logger.debug(f"_getRICVersion rslt {self.ricSystemInfo.get('rslt', '')} self.ricSystemInfo {self.ricSystemInfo}")
        return self.ricSystemInfo.get("rslt", "") == "ok"

    def _updateHwElemsInfo(self):
        hwElemsInfo = self.ricIF.cmdRICRESTURLSync("hwstatus")
        if hwElemsInfo.get("rslt", "") == "ok":
            self.ricHwElemsList = hwElemsInfo.get("hw", [])
            self.ricHwElemsInfoByIDNo = {}
            for el in self.ricHwElemsList:
                if "IDNo" in el:
                    self.ricHwElemsInfoByIDNo[el["IDNo"]] = el
            if self.DEBUG_GET_HW_ELEMS_INFO:
                logger.debug(f"_updateHwElemsInfo found {len(self.ricHwElemsList)} elems")

    def get_test_output(self) -> dict:
        return self.ricIF.getTestOutput()

    def _systemVersionGtEq(self, compareToVersion):
        if self.ricSystemInfo is None:
            return False
        versInfo = self.ricSystemInfo.get("SystemVersion","0.0.0")
        return version.parse(versInfo) >= version.parse(compareToVersion)

    def _subscribeToPubMessages(self, forceResubscribe: bool):
        versOk = self._systemVersionGtEq(self._minSysVersForSubscribeAPI)
        timeForSubscr = self.lastRICSerialMsgTime is None or time.time() > self.lastRICSerialMsgTime + self.maxTimeBetweenPubs
        resubscrReqd = self.lastSubscrReqMsgTime is None or time.time() > self.lastSubscrReqMsgTime + self.minTimeBetweenSubReqs
        if versOk and self._initComplete and self.subscribeRateHz != 0 and (forceResubscribe or (timeForSubscr and resubscrReqd)):
            # Subscribe for publication messages
            if self.DEBUG_SUBSCRIBE_TO_PUB_MSGS:
                logger.debug(f"Subscribe to published messages")
            self.ricIF.sendRICRESTCmdFrame('{"cmdName":"subscription","action":"update",' + \
                            '"pubRecs":[' + \
                                '{' + f'"name":"MultiStatus","rateHz":{self.subscribeRateHz},' + '}' + \
                                '{"name":"PowerStatus","rateHz":1.0},' + \
                                '{' + f'"name":"AddOnStatus","rateHz":{self.subscribeRateHz}' + '}' + \
                            ']}')
            self.lastSubscrReqMsgTime = time.time()

    def _unsubscribeFromPubMessages(self):
        # Send unsubscribe request
        self.ricIF.sendRICRESTCmdFrame('{"cmdName":"subscription","action":"update",' + \
            '"pubRecs":[' + \
                '{"name":"MultiStatus","rateHz":0},' + \
                '{"name":"PowerStatus","rateHz":0},' + \
                '{"name":"AddOnStatus","rateHz":0}' + \
            ']}')
        # Allow message to be sent
        time.sleep(0.5)

    def _onReconnect(self):
        self._subscribeToPubMessages(True)

    def _sendFileProgressAdapter(self, fileSize: int, bytesSent: int,
                progress_callback: Callable[[int, int, 'RICInterface'], bool] = None):
        if self._sendFileProgressCB:
            return self._sendFileProgressCB(fileSize, bytesSent)
        return True

    def send_file(self, filename: str,  
                progress_callback: Callable[[int, int], bool] = None,
                file_dest:str = "fs") -> bool:
        self._sendFileProgressCB = progress_callback
        return self.ricIF.sendFile(filename, self._sendFileProgressAdapter, file_dest)

    def _playMP3ProgressAdapter(self, fileSize: int, bytesSent: int,
                progress_callback: Callable[[int, int, 'RICInterface'], bool] = None):
        if self._playMP3ProgressCB:
            return self._playMP3ProgressCB(fileSize, bytesSent)
        return True

    def play_mp3(self, filename: str, 
                progress_callback: Callable[[int, int], bool] = None) -> bool:
        self._playMP3ProgressCB = progress_callback
        return self.ricIF.streamSoundFile(filename, "streamaudio", self._playMP3ProgressAdapter)

    def get_file_list(self) -> List[str]:
        result = self.ricIF.cmdRICRESTURLSync("filelist", timeOutSecs=5)
        if result.get("rslt", "") == "ok":
            return result.get("files", [])
        return []

    def delete_file(self, filename: str) -> bool:
        result = self.ricIF.cmdRICRESTURLSync(f"filedelete/local/{filename}", timeOutSecs=5)
        return result.get("rslt", "") == "ok"

    def rgb_to_hex(self, rgb: Tuple[int, int, int]) -> str:
        """Convert an RGB color to its hexadecimal representation."""
        return "{:02x}{:02x}{:02x}".format(int(rgb[0]), int(rgb[1]), int(rgb[2]))

    def led_id_mapping(self, id, is_from_color_picker, add_on_who_am_i='LEDeye'):

        if add_on_who_am_i == 'LEDeye':
            # Map LED position id to code id
            # The order starting from the top id is: 6 5 4 3 2 1 0 11 10 9 8 7
            MAP = [6, 5, 4, 3, 2, 1, 0, 11, 10, 9, 8, 7]
            if is_from_color_picker:
                return MAP[(id + 3) % 12]
            return MAP[id]
        
        elif add_on_who_am_i == 'LEDarm':
            MAP = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25]
        
        elif add_on_who_am_i == 'LEDfoot':
            MAP = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
            return MAP[id]

        return id

    def _get_hue_color(self, r, g, b):
        maxVal = max(r, g, b)
        minVal = min(r, g, b)
        chroma = maxVal - minVal
        hue = 0
        if r >= g and r >= b:
            hue = (((g-b)/chroma)%6) * 60
        elif g >= b:
            hue = (((b-r)/chroma) + 2) * 60
        else:
            hue = (((r-g)/chroma) + 4) * 60
        if hue < 0:
            hue += 360
        return [hue, chroma]


'''
RICStreamHandler
'''
from typing import Callable
from .Exceptions import MartyTransferException
import logging
import os
import time

logger = logging.getLogger(__name__)

class RICStreamHandler:

    def __init__(self, ricInterface: 'RICInterface'):
        # Stream vars
        self._ricInterface = ricInterface
        self._streamId: int = None

    def streamSoundFile(self, fileName: str, targetEndpoint: str,
                progressCB: Callable[[int, int, 'RICInterface'], bool] = None) -> bool:
        '''
        Stream sound from the file system
        Args:
            fileName: The file to stream
            progressCB: callback used to indicate how stream is progressing, callable takes three params
                    which are bytesSent, totalBytes and the interface to RIC (of type RICInterface) and
                    returns a bool which should be True to continue the stream or False to abort
        Returns:
            True if operation succeeded
        Throws:
            OSError: operating system exceptions
            May throw other exceptions so include a general exception handler
        '''
        # Check validity
        try:
            fileStat = os.stat(fileName)
        except OSError as e:
            raise MartyTransferException("File not found")

        # Open the file
        with open(fileName, 'rb') as soundFile:

            # Setup file info 
            streamLength = fileStat.st_size
            streamType = "rtstream"
            streamName = os.path.basename(fileName)

            # Establish a maximum duration for streaming based on the assumption
            # that the file is encoded at at 10Kbits/s so a 1K file has 1s duration
            # with a 50% margin of error
            maxStreamDuration = streamLength / 1024 * 1.5

            # Send stream start message
            streamStartMsg = f'"cmdName":"ufStart","reqStr":"ufStart","fileType":"{streamType}","fileName":"{streamName}","endpoint":"{targetEndpoint}"'
            streamStartMsg += f',"fileLen":{streamLength}'
            resp = self._ricInterface.sendRICRESTCmdFrameSync('{' + streamStartMsg + '}')
            if resp.get("rslt","") != "ok":
                raise MartyTransferException("Stream start not acknowledged")
            self._ricInterface._streamStart()

            # Get stream details
            logger.debug(f"Stream started {resp}")
            self._streamId = resp.get("streamID", -1)
            maxBlockSize = resp.get("maxBlkSize", 1024)

            # Check streamId is valid
            if self._streamId < 0:
                raise MartyTransferException("Stream ID invalid")

            # File pos
            filePos = 0

            # Rate estimation
            msgRateWindow = []
            streamLastSoktoPos = 0
            intraMessageTime = 0.05
            lastMsgSentTime = 0
            lastMsgSent = False

            # Send stream data
            streamStartTime = time.time()
            while time.time() - streamStartTime < maxStreamDuration:

                # Wait for intra message time
                time.sleep(0.05 if intraMessageTime < 0.25 else intraMessageTime - 0.15)

                # # Reduce intra message time on each block to ensure we don't starve the connection
                # intraMessageTime -= 0.05

                # Check for sokto message
                isNewSokto, streamOkTo, streamClosed = self._ricInterface._streamGetLatest()
                if streamClosed:
                    # logger.debug("streamSoundFile stream closed")
                    break
                elif isNewSokto:
                    if streamOkTo >= streamLength:
                        # logger.debug(f"streamSoundFile sokto stream end reached")
                        break
                    else:
                        # logger.debug(f"streamSoundFile sokto not last message")
                        lastMsgSent = False
                    filePos = streamOkTo
                    if streamLastSoktoPos != filePos:
                        msgRateWindow.append((filePos, time.time()))
                        streamLastSoktoPos = filePos
                        if len(msgRateWindow) > 5:
                            msgRateWindow.pop(0)
                        msgRateWindowLen = len(msgRateWindow)
                        if msgRateWindowLen > 1:
                            timeBetweenOkTo = msgRateWindow[-1][1] - msgRateWindow[0][1]
                            bytesBetweenOkTo = msgRateWindow[-1][0] - msgRateWindow[0][0]
                            dataRate = bytesBetweenOkTo / timeBetweenOkTo
                            intraMessageTime = maxBlockSize / dataRate
                            if intraMessageTime > 1.0:
                                intraMessageTime = 1.0
                            # Debug
                            # logger.debug(f"streamSoundFile recalc {bytesBetweenOkTo} between {round(timeBetweenOkTo,2)}s rate {round(dataRate,0)}bytesps intraMsg {round(intraMessageTime,2)}s")

                # Progress and check for abort
                if self._sendStreamProgressCheckAbort(progressCB, filePos, streamLength):
                    # logger.debug("streamSoundFile progressCB aborted")
                    break

                # Read block
                soundFile.seek(filePos)
                fileBlock = soundFile.read(maxBlockSize)

                # Stream data
                if len(fileBlock) > 0:
                    if not self._ricInterface.sendRICRESTFileBlock(self._streamId.to_bytes(1, 'big') +
                            filePos.to_bytes(3, 'big') + fileBlock):
                        break

                    # Bump file pos
                    filePos += len(fileBlock)
                else:
                    # logger.debug(f"streamSoundFile end of file")
                    lastMsgSent = True
                    lastMsgSentTime = time.time()

                # Check for timeout on last frame sent
                if lastMsgSent and time.time() - lastMsgSentTime > 2:
                    # logger.debug(f"streamSoundFile timeout")
                    break

            # End frame
            resp = self._ricInterface.sendRICRESTCmdFrameSync('{' + f'"cmdName":"ufEnd","reqStr":"ufEnd","streamId":"{self._streamId}"' + '}')
            if resp.get("rslt","") != "ok":
                return False
        return True

    def _sendStreamProgressCheckAbort(self, progressCB: Callable[[int, int, 'RICInterface'], bool], 
                    currentPos: int, fileSize: int) -> bool:
        if not progressCB:
            return False
        if not self._ricInterface.isOpen():
            return True
        if not progressCB(currentPos, fileSize, self):
            self._ricInterface.sendRICRESTCmdFrameSync('{' + f'"cmdName":"ufCancel", "streamId":"{self._streamId}"' + '}')
            return True
        return False




import sys    
import logging

import pathlib
cur_path = pathlib.Path(__file__).parent.resolve()
sys.path.insert(0, str(cur_path.parent.parent.resolve()))

from martypy import Marty
import time

# Setup logging
logging.basicConfig(format='%(levelname)s: %(asctime)s %(funcName)s(%(lineno)d) -- %(message)s', level=logging.DEBUG)
logger = logging.getLogger("testColourSensor")
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)

mm = Marty("usb")

# time.sleep(2)

for i in range(100):
    try:
        onGroundL = mm.foot_on_ground('left')
        onGroundR = mm.foot_on_ground('right')
        obstacleL = mm.foot_obstacle_sensed('left')
        obstacleR = mm.foot_obstacle_sensed('right')
        obstSensorL = mm.get_obstacle_sensor_reading('left')
        obstSensorR = mm.get_obstacle_sensor_reading('right')
        groundSensorL = mm.get_ground_sensor_reading('left')
        groundSensorR = mm.get_ground_sensor_reading('right')
        print(
            f"onGroundL={onGroundL!s:<5} onGroundR={onGroundR!s:<5} | "
            f"obstacleL={obstacleL!s:<5} obstacleR={obstacleR!s:<5} || "
            f"obstSensorL={obstSensorL:5} obstSensorR={obstSensorR:5} | "
            f"groundSensorL={groundSensorL:5} groundSensorR={groundSensorR:5}"
        )
    except Exception as e:
        logger.exception(f"Failed to get addon values")        
    time.sleep(.2)

# try:
# mm = Marty("wifi://192.168.86.81")
# except Exception as excp:
#     print(f"Couldn't connect to marty")
#     exit()


# time.sleep(5)

mm.close()
